from django.db import models
from django.contrib.auth.models import AbstractUser
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import Q, QuerySet
from django.contrib.auth.models import Permission
from django.forms import ValidationError
from bloomerp.models.core import BloomerpModel, ApplicationField
from bloomerp.models.widgets import Widget
from django.utils.translation import gettext as _
from bloomerp.utils.models import model_name_plural_underline, get_detail_view_url
from django.urls import reverse, NoReverseMatch
from bloomerp.models.mixins import (
    AbsoluteUrlModelMixin,
    TimestampedModelMixin,
    ContentLayoutModelMixin,
    StringSearchModelMixin,
    UserStampedModelMixin
)
from typing import Self

# ---------------------------------
# User Model
# ---------------------------------
class AbstractBloomerpUser(
    AbstractUser,
    AbsoluteUrlModelMixin,
    ):
    class Meta:
        abstract = True


    string_search_fields = ['first_name+last_name', 'username']
    allow_string_search = True

    avatar = models.ImageField(null=True, blank=True, upload_to="users/", help_text=_("The user's avatar"))

    # ------------------------------------------------
    # User Preferences
    # ------------------------------------------------
    FILE_VIEW_PREFERENCE_CHOICES = [
        ("card", "Card View"),
        ("list", "List View"),
    ]

    DATE_VIEW_PREFERENCE_CHOICES = [
        ("d-m-Y", "Day-Month-Year (15-08-2000)"),
        ("m-d-Y", "Month-Day-Year (08-15-2000)"),
        ("Y-m-d", "Year-Month-Day (2000-08-15)"),
    ]

    DATETIME_VIEW_PREFERENCE_CHOICES = [
        ("d-m-Y H:i", "Day-Month-Year Hour:Minute (15-08-2000 12:30)"),
        ("m-d-Y H:i", "Month-Day-Year Hour:Minute (08-15-2000 12:30)"),
        ("Y-m-d H:i", "Year-Month-Day Hour:Minute (2000-08-15 12:30)"),
    ]

    file_view_preference = models.CharField(
        max_length=20, default="card", choices=FILE_VIEW_PREFERENCE_CHOICES
    )

    date_view_preference = models.CharField(
        max_length=20, default="d-m-Y", choices=DATE_VIEW_PREFERENCE_CHOICES, help_text=_("The date format to be used in the application")
    )

    datetime_view_preference = models.CharField(
        max_length=20, default="d-m-Y H:i", choices=DATETIME_VIEW_PREFERENCE_CHOICES, help_text=_("The datetime format to be used in the application")
    )

    #sidebar_preference = models.JSONField(default=dict)
    

    def __str__(self):
        return self.username

    def is_employee(self):
        return hasattr(self, "employee")

    def get_content_types_for_user(self, permission_types:list[str]=["view"]) -> QuerySet[ContentType]:
        """
        Get all content types the user has permissions for based on the provided permission types.
        Permission types are the prefixes of the permission codenames, e.g. 'view', 'add', 'change', 'delete'.
        """
        if self.is_superuser:
            return ContentType.objects.all()

        # Build the query for filtering permissions based on the provided types
        permission_filters = Q()
        for perm_type in permission_types:
            permission_filters |= Q(codename__startswith=perm_type + "_")

        # Get all permissions for the user, including those via groups
        user_permissions = self.user_permissions.filter(
            permission_filters
        ) | Permission.objects.filter(permission_filters, group__user=self)

        # Get the content types for all permissions the user has
        content_types = ContentType.objects.filter(
            permission__in=user_permissions
        ).distinct()

        return content_types

    def get_list_view_preference_for_model(self, model) -> QuerySet:
        """
        Get the list view preference for the provided model.
        """
        content_type = ContentType.objects.get_for_model(model)
        return UserListViewPreference.objects.filter(user=self, application_field__content_type=content_type)

    @property
    def accessible_content_types(self) -> QuerySet:
        '''
        Property that returns all content types the user has view access to.
        '''
        return self.get_content_types_for_user(permission_types=["view"])

    
    def latest_bookmarks(self) -> QuerySet:
        '''
        Property that returns the latest bookmarks for the user.
        '''
        return Bookmark.objects.filter(user=self).order_by('-datetime_created')[:5]

    def workspaces(self) -> QuerySet:
        '''
        Method that returns the workspaces for the user.
        '''
        return Workspace.objects.filter(user=self, content_type=None)


class User(AbstractBloomerpUser, StringSearchModelMixin):
    class Meta(BloomerpModel.Meta):
        db_table = "auth_user"
        swappable = "AUTH_USER_MODEL"


# ---------------------------------
# User Detail View Preference Model
# ---------------------------------
from django.db.models import BooleanField, Case, When, Subquery, OuterRef
class UserDetailViewPreference(
    models.Model,
    ):
    POSITION_CHOICES = [
        ('LEFT','Left'), ('CENTER','Center'),('RIGHT','Right')
    ]

    allow_string_search = False

    user = models.ForeignKey(User, on_delete=models.CASCADE,related_name = 'detail_view_preference')
    application_field = models.ForeignKey(ApplicationField, on_delete=models.CASCADE)
    position = models.CharField(max_length=10, choices=POSITION_CHOICES)

    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_user_detail_view_preference'
        unique_together = ('user', 'application_field')

    def get_application_fields_info(content_type_id, user):
        
        is_used_subquery = Subquery(
            UserDetailViewPreference.objects.filter(
                user=user,
                application_field=OuterRef('pk')
            ).values('application_field').annotate(is_used=Case(
                When(pk=OuterRef('pk'), then=True),
                default=True,
                output_field=BooleanField()
            )).values('is_used')[:1]
        )

        position = UserDetailViewPreference.objects.filter(
            user=user,
            application_field=OuterRef('pk')
        ).values('position')[:1]

        application_fields_info = ApplicationField.objects.filter(
            content_type_id=content_type_id
        ).annotate(
            is_used=is_used_subquery,
            position=position
        ).values(
            'field',
            'id',
            'is_used',  
            'position'
        )

        return list(application_fields_info)


    @classmethod
    def generate_default_for_user(cls, user: User, content_type: ContentType) -> QuerySet[Self]:
        '''
        Method that generates default detail view preference for a user.
        '''
        application_fields = ApplicationField.objects.filter(content_type=content_type)
        
        # Exclude some application fields
        application_fields = application_fields.exclude(
            field_type__in=['ManyToManyField', 'OneToManyField']
        )

        # Exclude some more fields
        application_fields = application_fields.exclude(
            field='id'
        )
        
        
        for application_field in application_fields:
            preference, created = UserDetailViewPreference.objects.get_or_create(
                user=user,
                application_field=application_field,
                position='LEFT'
            )
            
        return UserDetailViewPreference.objects.filter(user=user, application_field__content_type=content_type)

# ---------------------------------
# User List View Preference Model
# ---------------------------------
class UserListViewPreference(models.Model):
    allow_string_search = False
    
    user = models.ForeignKey(User, on_delete=models.CASCADE,related_name = 'list_view_preference')
    application_field = models.ForeignKey(ApplicationField, on_delete=models.CASCADE)

    @property
    def field_name(self):
        return (self.application_field.field).replace('_', ' ').title()

    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_user_list_view_preference'
        unique_together = ('user', 'application_field')


    @classmethod
    def generate_default_for_user(cls, user: User, content_type: ContentType) -> QuerySet[Self]:
        '''
        Method that generates default list view preference for a user.
        '''
        application_fields = ApplicationField.objects.filter(content_type=content_type)
        
        # Exclude some application fields
        application_fields = application_fields.exclude(
            field_type__in=['ManyToManyField', 'OneToManyField']
        )

        # Exclude some more fields
        application_fields = application_fields.exclude(
            field__in=['id', 'created_by', 'updated_by', 'datetime_created', 'datetime_updated']
        )
        
        # Only take the first 5 fields
        for application_field in application_fields[:5]:
            preference, created = UserListViewPreference.objects.get_or_create(
                user=user,
                application_field=application_field
            )
            
        return UserListViewPreference.objects.filter(user=user, application_field__content_type=content_type)

# ---------------------------------
# Bookmark Model
# ---------------------------------
class Bookmark(models.Model):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = "bloomerp_bookmark"
    
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    object : models.Model = GenericForeignKey("content_type", "object_id")

    datetime_created = models.DateTimeField(auto_now_add=True)

    allow_string_search = False

    def __str__(self) -> str:
        return f"Bookmark for {self.content_type} with ID {self.object_id}"

    def get_absolute_url(self):
        try:
            return self.object.get_absolute_url()
        except:
            model = self.object._meta.model
            detail_view_url = get_detail_view_url(model)
            return reverse(detail_view_url, args=[self.object.pk])
            
# ---------------------------------
# Bloomerp Comment Model
# ---------------------------------
class Comment(
    TimestampedModelMixin,
    UserStampedModelMixin,
    models.Model,
):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_comment'

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.CharField(max_length=36) # In order to support both UUID and integer primary keys
    content_object = GenericForeignKey("content_type", "object_id")
    content = models.TextField()

    allow_string_search = False

    def __str__(self):
        return f"{self.content} - {self.created_by} - {self.datetime_created}"

# ---------------------------------
# Bloomerp Todo Model
# ---------------------------------
class Todo(BloomerpModel):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_todo'

    avatar = None
    assigned_to = models.ForeignKey(User, on_delete=models.CASCADE, related_name='todos')
    requested_by = models.ForeignKey(User, null=True, blank=True, on_delete=models.CASCADE, related_name='requested_todos', help_text=_("The user who requested the todo"))

    required_by = models.DateField(
        null=True, blank=True,
        help_text=_("The date by which the todo is required")
        )
    priority = models.IntegerField(
        help_text=_("The priority of the todo"), 
        choices=[(1, 'Low'), (2, 'Medium'), (3, 'High')],
        default=2
        )

    title = models.CharField(max_length=255, help_text=_("The name of the todo"))
    content = models.TextField(blank=True, null=True)

    is_completed = models.BooleanField(default=False)
    datetime_completed = models.DateTimeField(null=True, blank=True)

    # For if the todo is related to a model
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True, blank=True)
    object_id = models.CharField(max_length=36, null=True, blank=True) # In order to support both UUID and integer primary keys
    content_object = GenericForeignKey("content_type", "object_id")


    allow_string_search = False # Do not allow string search for todos (we dont want to-do's to be searchable in the search bar)
    string_search_fields = ['content'] # Allow string search for content

    @property
    def priority_string(self):
        if self.priority == 1:
            return "Low"
        elif self.priority == 2:
            return "Medium"
        elif self.priority == 3:
            return "High"

    def __str__(self):
        return self.title


    def clean(self):
        errors = {}
        from django.utils import timezone
        from django.core.exceptions import ObjectDoesNotExist

        # Set the datetime completed to None if the todo is not completed
        if self.is_completed and not self.datetime_completed:
            self.datetime_completed = timezone.now()
        elif not self.is_completed:
            self.datetime_completed = None


        if self.content_type and self.object_id:
            try:
                self.content_object  # Triggers a lookup
            except ObjectDoesNotExist:
                errors['content_object'] = _("The related object does not exist")

        if errors:
            raise ValidationError(errors)

        return super().clean()

class Link(
    AbsoluteUrlModelMixin,
    StringSearchModelMixin,
    models.Model,
):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_link'

    LEVEL_TYPES = [
        ('DETAIL', 'Detail'),
        ('LIST', 'List'),
        ('APP', 'App'),
    ]

    name = models.CharField(max_length=255)
    url = models.CharField(max_length=255, help_text=_("The name of the URL pattern for the link. Can either be a Django URL name or a full URL (absolute)."))
    level = models.CharField(max_length=255, choices=LEVEL_TYPES, help_text=_("The level of the link."))
    is_absolute_url = models.BooleanField(
        default=True,
        help_text=_("Signifies whether the URL is a normal (absolute) URL or a Django URL name. Set to True when using a normal URL.")
        )
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True, blank=True, help_text=_("The content type for the link. Required for detail and list levels."))
    description = models.TextField(null=True, blank=True, help_text=_("Description of the link"))
    
    string_search_fields = ['name', 'url', 'description']
    allow_search = False

    def __str__(self):
        return self.name
    
    def clean(self):
        errors = {}

        # Check if the level type is valid
        if self.level not in [level[0] for level in self.LEVEL_TYPES]:
            errors['level'] = _("Invalid level type")

        # Check if the content type is set for the detail level or list level
        if self.level in ['DETAIL', 'LIST'] and not self.content_type:
            errors['content_type'] = _("Content type must be set for the detail or list level")
        
        # Check if the content type is not set for the app level
        if self.level == 'APP' and self.content_type:
            errors['content_type'] = _("Content type must not be set for the app level")
        
        # If not an absolute URL, check if the reverse works
        if not self.is_absolute_url:
            try:
                reverse(self.url_name)
            except:
                errors['url_name'] = _("The URL name does not match any URL pattern. Set is_absolute_url to True or give a valid URL name.")

        # If the URL is not a Django URL name, and has been saved before, it cannot be changed
        if not self.is_absolute_url and self.pk:
            old_link = Link.objects.get(pk=self.pk)
            if old_link.url != self.url:
                errors['url'] = _("The URL cannot be changed for a link that is not an absolute URL.")

        # Make sure the Link is valid
        if not self.is_valid():
            errors['url'] = _("The URL is not valid.")

        if errors:
            raise ValidationError(errors)

        return super().clean()
    
    def create_default_list_links_for_content_type(content_type: ContentType, output='dict') -> dict:
        """
        Function that creates default links for a content type.

        Output can be either a dictionary or a list of links.

        The default links are:
            - List link
            - App link
            - Create link
            - Bulk upload link
        """
        from bloomerp.utils.models import (
            get_list_view_url,
            get_create_view_url,
            get_model_dashboard_view_url,
            get_bulk_upload_view_url
        )
        # Get the model name
        model = content_type.model_class()

        # Create the list view link
        list_link, created = Link.objects.get_or_create(
            name = _(f"{model._meta.verbose_name.title()} list"),
            url = get_list_view_url(model),
            level = 'LIST',
            content_type = content_type,
            description = _(f"List view for {model._meta.verbose_name}"),
            is_absolute_url = False
        )        
        
        # Create the app link
        app_link, created = Link.objects.get_or_create(
            name = _(f"{model._meta.verbose_name.title()} App"),
            url = get_model_dashboard_view_url(model),
            level = 'APP',
            content_type = content_type,
            description = _(f"App view for {model._meta.verbose_name}"),
            is_absolute_url = False
        )

        # Create the create link
        create_link, created = Link.objects.get_or_create(
            name = _(f"Create {model._meta.verbose_name.title()}"),
            url = get_create_view_url(model),
            level = 'LIST',
            content_type = content_type,
            description = _(f"Create view for {model._meta.verbose_name}"),
            is_absolute_url = False
        )

        # Create the bulk upload link
        bulk_upload_link, created = Link.objects.get_or_create(
            name = _(f"Bulk Upload {model._meta.verbose_name.title()}"),
            url = get_bulk_upload_view_url(model),
            level = 'LIST',
            content_type = content_type,
            description = _(f"Bulk upload view for {model._meta.verbose_name}"),
            is_absolute_url = False
        )

        return {
            'list_link': list_link,
            'app_link': app_link,
            'create_link': create_link,
            'bulk_upload_link': bulk_upload_link
        }

    def create_default_detail_links_for_content_type(content_type: ContentType) -> dict:
        '''
        Function that creates default links for a content type.
        The default links are:
            - Detail link
            - Update link
        '''
        from bloomerp.utils.models import get_detail_view_url, get_update_view_url
        model = content_type.model_class()

        # Create the detail view link
        detail_link, created = Link.objects.get_or_create(
            name = _(f"{model._meta.verbose_name.title()} detail"),
            url = get_detail_view_url(model),
            level = 'DETAIL',
            content_type = content_type,
            description = _(f"Detail view for {model._meta.verbose_name}"),
            is_absolute_url = False
        )

        # Create the update view link
        update_link, created = Link.objects.get_or_create(
            name = _(f"Update {model._meta.verbose_name.title()}"),
            url = get_update_view_url(model),
            level = 'DETAIL',
            content_type = content_type,
            description = _(f"Update view for {model._meta.verbose_name}"),
            is_absolute_url = False
        )

        return {
            'detail_link': detail_link,
            'update_link': update_link
        }

    def get_list_links_for_content_types(content_types: QuerySet, name=None) -> list[dict]:
        '''
        Function that returns the links for a particular content type. 
        If a query is provided, the links will be additionally filtered by name.

        Args:
            content_types: QuerySet[ContentType]
            name: str

        Returns:
            [{'model_name': str, 'links': QuerySet[Link]}]
        '''
        links_info = []

        for content_type in content_types:
            links = Link.objects.filter(content_type=content_type, level='LIST')
            if name:
                links = links.filter(name__icontains=name)
            
            links_info.append({
                'model_name': content_type.model_class()._meta.verbose_name,
                'links': links
            })

        return links_info
    
    def detail_view_tab_links(content_type: ContentType) -> QuerySet:
        '''
        Method that returns the detail view tab links for a content type.
        '''
        qs = Link.objects.filter(content_type=content_type, level='DETAIL') 
        
        for link in qs:
            if link.number_of_args() > 1:
                # Exclude links that require more than one argument
                qs = qs.exclude(pk=link.pk)

        return qs

    @property
    def model_name(self) -> str:
        '''
        Property that returns the model name of the link.
        '''
        if self.content_type:
            return self.content_type.model_class()._meta.verbose_name
        else:
            return None
        
    def requires_args(self) -> bool:
        '''
        Method that checks if the link requires arguments.
        '''
        if self.is_absolute_url:
            return False
        else:
            try:
                reverse(self.url)
                return False
            except NoReverseMatch:
                return True

    def to_absolute_url(self) -> str:
        from django.urls import reverse
        if self.is_absolute_url:
            return self.url
        else:
            try:
                return reverse(self.url)
            except:
                pass

    def get_args(self) -> list:
        '''
        Returns the arguments required for the link.
        '''
        from django.urls import get_resolver
        if self.is_absolute_url:
            return []
        else:
            try:
                resolver = get_resolver()
                return resolver.reverse_dict[self.url][0][0][1]
            except Exception as e:
                return []

    def number_of_args(self) -> int:
        '''Returns the number of args required for the link.'''
        return len(self.get_args())

    def is_external_url(self) -> bool:
        '''
        Method that checks if the link is an external URL.
        External URLs are URLs that are not part of the application.

        Example:
            - https://www.google.com
            - https://www.example.com

        So will return True if the link has www. or http in it.
        
        '''
        if not self.is_absolute_url:
            return False
        else:
            return 'www.' in self.url or 'http' in self.url

    def is_valid(self) -> bool:
        '''
        Method that checks if the link is valid.
        '''
        try:
            if self.is_absolute_url:
                return True
            else:
                from django.urls import get_resolver
                resolver = get_resolver()
                resolver.reverse_dict[self.url]
                return True
        except:
            return False

class UserDetailViewTab(
    AbsoluteUrlModelMixin,

    models.Model
    ):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_user_detail_view_tab'
        unique_together = ('user','link')

    user = models.ForeignKey(User, on_delete=models.CASCADE)
    link = models.ForeignKey(Link, help_text=_("The link to be displayed in the detail view tab"), on_delete=models.CASCADE)

    allow_string_search = False

    def get_detail_view_tabs(user:User, content_type:ContentType) -> QuerySet[Self]:
        '''
        Returns the detail view tabs for the user and content type.
        '''
        qs = UserDetailViewTab.objects.filter(user=user, link__content_type=content_type, link__level='DETAIL')
        return qs

    @classmethod
    def generate_default_for_user(cls, user: User, content_type: ContentType) -> QuerySet[Self]:
        '''
        Method that generates default detail view tabs for a user.
        '''
        links : QuerySet[Link] = Link.detail_view_tab_links(content_type)
        
        for link in links:
            UserDetailViewTab.objects.get_or_create(
                user=user,
                link=link
            )
            
        return UserDetailViewTab.objects.filter(user=user, link__content_type=content_type)

    def __str__(self):
        return str(self.user) + ' ' + str(self.link.name)

    def clean(self):
        errors = {}

        # Check if the link is a detail link
        if self.link.level != 'DETAIL':
            errors['link'] = _("Link must be a detail link")

        if self.link.number_of_args() > 1:
            errors['link'] = _("Link can only have one argument (pk) for a detail view tab")

        if errors:
            raise ValidationError(errors)

        return super().clean()


def get_default_workspace():
    links = Link.objects.all()
    widgets = Widget.objects.all()

    return {
        "content" : [
            {
                "type": "header",
                "data": {"text": "Welcome to your workspace"},
                "size" : 12
            },
            {
                "type": "text",
                "data": {"text": "This is your workspace. You can add widgets, links, and other content here."},
                "size" : 12
            },
            {
                "type": "header",
                "data": {"text": "Example of widget"},
                "size" : 12
            },
            
            {
                "type": "header",
                "data": {"text": "Example of link"},
                "size" : 12
            },
            {
                "type": "link",
                "data": {"link_id": links.first().pk},
                "size" : 12
            },
            {
                "type": "header",
                "data": {"text": "Example of link list"},
                "size" : 12
            },
            {
                "type": "link_list",
                "data": {"links": [links.first().pk, links.last().pk]},
                "size" : 12
            }
        ]
    }

class Workspace(
    AbsoluteUrlModelMixin,
    models.Model,
    ):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_workspace'

    WORKSPACE_TYPES = [
        "header",
        "text",
        "widget",
        "link",
        "link_list",
        "query_filter",
        "query_filter_list"
    ]

    user = models.ForeignKey(User, on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.SET_NULL, null=True, blank=True, help_text=_("The content type related to the workspace"))
    name = models.CharField(max_length=255, help_text=_("The name of the workspace"))
    content = models.JSONField(help_text=_("The content of the workspace"), default=get_default_workspace)

    def __str__(self):
        return self.name

    def clean(self):
        # Check if the content is valid
        errors = {}
        if not self.content.get('content'):
            errors['content'] = _("Content must be provided")

        for item in self.content['content']:
            if item['type'] not in self.WORKSPACE_TYPES:
                errors['content'] = _("Invalid content type")

        # Check if the content is valid for links (link must have a valid link_id)
        for item in self.content['content']:
            if item['type'] == 'link':
                if not item['data'].get('link_id'):
                    errors['content'] = _("Link must have a link_id")
                else:
                    try:
                        Link.objects.get(pk=item['data']['link_id'])
                    except:
                        errors['content'] = _("Link does not exist")

        # Check if the content is valid for link lists (link list must have a list of valid link ids)
        for item in self.content['content']:
            if item['type'] == 'link_list':
                if not item['data'].get('links'):
                    errors['content'] = _("Link list must have a list of links")
                else:
                    for link_id in item['data']['links']:
                        try:
                            Link.objects.get(pk=link_id)
                        except:
                            errors['content'] = _(f"Link with id {link_id} does not exist")

        # Check if the content is valid for widgets (widget must have a valid widget_id)
        for item in self.content['content']:
            if item['type'] == 'widget':
                if not item['data'].get('widget_id'):
                    errors['content'] = _("Widget must have a widget_id")
                else:
                    try:
                        Widget.objects.get(pk=item['data']['widget_id'])
                    except:
                        errors['content'] = _("Widget does not exist")
        
        if errors:
            raise ValidationError(errors)
        
        return super().clean()
    
    def save(self, *args, **kwargs):
        # Clean the data before saving
        self.clean()
        return super().save(*args, **kwargs)
    
    def remove_links_from_content(self, links: list[Link]):
        '''
        Method that removes links from the workspace content.
        '''
        content = self.content
        new_content = []

        for item in content['content']:
            if item['type'] == 'link':
                if item['data']['link_id'] not in [link.pk for link in links]:
                    new_content.append(item)
            elif item['type'] == 'link_list':
                new_links = []
                for link_id in item['data']['links']:
                    if link_id not in [link.pk for link in links]:
                        new_links.append(link_id)
                item['data']['links'] = new_links
                if new_links:
                    new_content.append(item)
            else:
                new_content.append(item)
        
        self.content = {'content': new_content}
        self.save()

    def remove_stale_links_content(self):
        '''
        Removes stale links from the workspace content.
        Stale links are links that do not exist in the database.
        '''
        content = self.content
        new_content = []

        for item in content['content']:
            if item['type'] == 'link':
                try:
                    Link.objects.get(pk=item['data']['link_id'])
                    new_content.append(item)
                except:
                    pass
            elif item['type'] == 'link_list':
                new_links = []
                for link_id in item['data']['links']:
                    try:
                        Link.objects.get(pk=link_id)
                        new_links.append(link_id)
                    except:
                        pass
                if new_links:
                    item['data']['links'] = new_links
                    new_content.append(item)
            else:
                new_content.append(item)

        self.content = {'content': new_content}
        self.save()
    
    
    @staticmethod
    def create_default_content_type_workspace(
        user: User,
        content_type: ContentType,
        commit: bool = True
    ) -> Self:
        '''
        Function that creates a default workspace for a user and content type.
        '''
        # Get link objects
        qs = Link.objects.filter(content_type=content_type)
        links = []

        for link in qs:
            if not link.requires_args():
                links.append(link.pk)

        content = {
            'content': [
                {
                    'type': 'header',
                    'data': {'text': f'Quick links for {content_type.model_class()._meta.verbose_name_plural}'},
                    'size': 12
                },
                {
                    'type': 'link_list',
                    'data': {'links': links},
                    'size': 12
                }
            ]
        }

        if commit:
            workspace, created = Workspace.objects.get_or_create(
                user=user,
                content_type=content_type,
                name=content_type.model_class()._meta.verbose_name_plural,
                content=content
            )
        else:
            workspace = Workspace(
                user=user,
                content_type=content_type,
                name=content_type.model_class()._meta.verbose_name_plural,
                content=content
            )

        return workspace

    @staticmethod
    def create_default_workspace(
        user: User,
        commit: bool = True
    ) -> Self:
        '''
        Function that creates a default workspace for a user.
        '''
        # Get accessible content types for the user
        content_types = user.get_content_types_for_user(['view'])

        # Get link objects
        qs = Link.objects.filter(content_type__in=content_types, level='LIST')
        links = []
        for link in qs:
            if not link.requires_args():
                links.append(link)



        content = {
            'content': [
                {
                    'type': 'header',
                    'data': {'text': f'Welcome to your Bloomerp workspace'},
                    'size': 12
                },
                {
                    'type': 'text',
                    'data': {'text': f'You can add widgets, links, and other content here.'},
                    'size': 12
                },
                {
                    'type': 'header',
                    'data': {'text': 'Some quick links'},
                    'size': 12
                }
            ]
        }

        if links:
            content['content'].append({
                'type': 'link_list',
                'data': {'links': [link.pk for link in links]},
                'size': 12
            })

        if commit:
            workspace, created = Workspace.objects.get_or_create(
                user=user,
                name='Default Workspace',
                content=content
            )
        else:
            workspace = Workspace(
                user=user,
                name='Default Workspace',
                content=content
            )

        return workspace
    

# ---------------------------------
# AI Conversation Model
# ---------------------------------
import uuid
class AIConversation(BloomerpModel):
    class Meta:
        managed = True
        db_table = "bloomerp_ai_conversation"
        verbose_name = "AI conversation"
        verbose_name_plural = "AI conversations"

    CONVERSATION_TYPES = [
        ('sql', 'SQL'), 
        ('document_template', 'Document Template Generator'), 
        ('tiny_mce_content', 'TinyMCE Content Generator'), 
        ('bloom_ai', 'Bloom AI'),
        ('code', 'Code Generator')
    ]

    avatar = None
    title = models.CharField(max_length=255, default='AI Conversation')
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    user = models.ForeignKey('bloomerp.User', on_delete=models.CASCADE)
    conversation_history = models.JSONField(null=True, blank=True)
    conversation_type = models.CharField(max_length=20, choices=CONVERSATION_TYPES, default='bloom_ai')
    auto_named = models.BooleanField(default=False, help_text="Whether the conversation has been auto-named")


    allow_string_search = False
    string_search_fields = ['title']

    @property
    def number_of_messages(self):
        return len(self.conversation_history)
    
    def __str__(self):
        return self.title
    



from bloomerp.models import (
    DocumentTemplate, User, FileFolder,
    File)
from bloomerp.utils.pdf import generate_pdf
from django.db.models import Model
from django.core.files.base import ContentFile
from django.template import engines
from django.template.loader import render_to_string
from bloomerp.utils.pdf import PdfHandler


class DocumentController:
    '''
    Controller class for everything related to documents.
    '''

    def __init__(self, document_template : DocumentTemplate=None, user : User = None) -> None:
        self.document_template = document_template
        self.user = user

    def create_document(
            self,
            document_template:DocumentTemplate, 
            instance:Model, 
            free_variables: dict=None,
            persist:bool=True
        ):
        ''' Creates a document for a particular template, using the model variable and the free variables:
            - Template : The document template
            - Instance : The model instance
            - Free variables : A dictionary of free variables

        '''
        data = {}

        if instance:
            data['object'] = instance
    	
        #Add free variable data
        if free_variables:
            data.update(free_variables)

        #Create metadata variable
        meta_data = {}
        meta_data['document_template'] = document_template.pk

        meta_data['signed'] = False
        
        #Format HTML       
        django_engine = engines["django"]
        temp = django_engine.from_string(document_template.template)
        formatted_html = temp.render(data)        

        # Get the styling
        # Check if the template has styling
        if not document_template.styling:
            styling = None
        else:
            styling = document_template.styling.styling

        # Check if document_template has a header
        if document_template.template_header:
            header = document_template.template_header
            header_url = header.header.path
            header_margin_bottom = header.margin_bottom
            header_margin_left = header.margin_left
            header_margin_right = header.margin_right
            header_margin_top = header.margin_top
            header_height = header.height
        else:
            header_url = None
            header_margin_bottom = 0
            header_margin_left = 0
            header_margin_right = 0
            header_margin_top = 0
            header_height = 0

        if document_template.page_orientation == 'landscape':
            is_landscape = True
        else:
            is_landscape = False

        document_bytes = generate_pdf(
            html_content=formatted_html,
            css_content=styling,
            # Header options
            header_url=header_url,
            header_margin_bottom=header_margin_bottom,
            header_margin_left=header_margin_left,
            header_margin_right=header_margin_right,
            header_margin_top=header_margin_top,
            header_height=header_height,
            # Page options
            title=document_template.name,
            page_size=document_template.page_size,
            page_margin=document_template.page_margin,
            footer=document_template.footer,
            is_landscape=is_landscape,
            include_page_numbers=document_template.include_page_numbers
        )

        content_file = ContentFile(document_bytes)
        
        if persist:
            file_object = File()
        
            #Save the file
            file_object.file.save(document_template.name + ' ' + str(instance) + '.pdf', content_file)

            file_object.name = document_template.name + ' ' + str(instance)
            file_object.content_object = instance

            # Add created by
            file_object.created_by = self.user
            file_object.updated_by = self.user
            

            #Save metadata
            file_object.meta = meta_data

            file_object.save()

            if document_template.save_to_folder:
                document_template.save_to_folder.files.add(file_object)
                

            return file_object
        else:
            return content_file
        
    def create_preview_document(
            self,
            document_template:DocumentTemplate,
            data:dict
        ) -> ContentFile:
        ''' Creates a preview document for a particular document_template, using the model variable and the free variables:
            - Template : The document document_template
            - Data : A dictionary of free variables
        '''
        #Format HTML
        django_engine = engines["django"]
        temp = django_engine.from_string(document_template.template)
        formatted_html = temp.render(data)
        

        # Check if the document_template has styling
        if not document_template.styling:
            styling = None
        else:
            styling = document_template.styling.styling

        # Check if document_template has a header
        if document_template.template_header:
            header = document_template.template_header
            header_url = header.header.path
            header_margin_bottom = header.margin_bottom
            header_margin_left = header.margin_left
            header_margin_right = header.margin_right
            header_margin_top = header.margin_top
            header_height = header.height
        else:
            header_url = None
            header_margin_bottom = 0
            header_margin_left = 0
            header_margin_right = 0
            header_margin_top = 0
            header_height = 0

        if document_template.page_orientation == 'landscape':
            is_landscape = True
        else:
            is_landscape = False

        document_bytes = generate_pdf(
            html_content=formatted_html,
            css_content=styling,
            # Header options
            header_url=header_url,
            header_margin_bottom=header_margin_bottom,
            header_margin_left=header_margin_left,
            header_margin_right=header_margin_right,
            header_margin_top=header_margin_top,
            header_height=header_height,
            # Page options
            title=document_template.name,
            page_size=document_template.page_size,
            page_margin=document_template.page_margin,
            footer=document_template.footer,
            is_landscape=is_landscape,
        )

        content_file = ContentFile(document_bytes)

        return content_file

    def sign_pdf(
            self, 
            file : File, 
            signature_bytes : bytes
            ) -> File:
        '''
        Function that will sign a pdf file, using signature bytes.
        '''
        file_path = file.file.path

        # Create metadata variable
        meta_data = {}

        # If a user is present, add the user to the metadata
        if self.user:
            meta_data['user'] = self.user.pk

        # Check if the file object has a document template
        if 'document_template' in file.meta:
            meta_data['document_template'] = file.meta['document_template']

        # Add signed equals to true
        meta_data['signed'] = True

        # Create a PdfHandler object
        handler = PdfHandler(file_path)

        #Sign the actual document and retreive the bytes
        document_bytes = handler.sign_pdf(
            signature_path=None,
            signature_bytes= signature_bytes
        )

        # Create a content file
        content_file = ContentFile(document_bytes)
        
        # Create a file object
        signed_file_obj = File()

        #Save the file
        signed_file_obj.file.save(file.name + '- signed' '.pdf', content_file)

        signed_file_obj.content_object = file.content_object

        signed_file_obj.meta = meta_data
        signed_file_obj.name = file.name + '- signed.pdf' 
        signed_file_obj.save()
        
        # Update the original file with the signed file id
        file.meta = {'signed_file_id': str(signed_file_obj.pk)}
        file.save()

        return signed_file_obj
        







# Generated by Django 5.1.1 on 2024-10-06 17:57

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0032_comment_created_by_comment_updated_by_and_more"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserAppDashboard",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "layout",
                    models.JSONField(
                        blank=True,
                        help_text="The layout of the app dashboard",
                        null=True,
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "dashboard",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="app_dashboard",
                        to="bloomerp.userdashboardview",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_user_app_dashboard",
                "managed": True,
                "unique_together": {("user", "content_type")},
            },
        ),
    ]




from django.test import TestCase, RequestFactory
from bloomerp.models import Link, ContentType, User
from bloomerp.components.search_results import search_results
from django.urls import reverse

class SearchResultsTests(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='testuser', password='12345', is_superuser=True)
        self.content_type = ContentType.objects.create(app_label='test_app', model='testmodel')
        Link.objects.create(name='Test Link', url='test/url', level='LIST', content_type=self.content_type)
        self.url = reverse('components_search_results')


    def test_no_query_provided(self):
        request = self.factory.get(self.url)
        request.user = self.user
        response = search_results(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'No query provided', response.content)

    def test_list_level_links(self):
        request = self.factory.get(self.url, {'search_results_query': '/testmodel'})
        request.user = self.user
        response = search_results(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'Test Link', response.content)

    def test_app_level_links(self):
        Link.objects.create(name='App Link', url='app/url', level='APP')
        request = self.factory.get(self.url, {'search_results_query': '//A'})
        request.user = self.user
        response = search_results(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'App Link', response.content)

    def test_model_string_search(self):
        # Assuming your model has a string_search method
        request = self.factory.get(self.url, {'search_results_query': 'test'})
        request.user = self.user
        response = search_results(request)
        self.assertEqual(response.status_code, 200)
        # Add more assertions based on your model's string_search results


# Generated by Django 5.1.1 on 2024-09-17 20:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0006_alter_order_options_alter_product_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="order",
            name="customer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="orders",
                to="test_app.customer",
            ),
        ),
    ]



from django.test import TestCase, RequestFactory
from django.urls import reverse
from bloomerp.models import User, Bookmark, ContentType, File, BloomerpModel, Link
from bloomerp.components.bookmark import bookmark
from bloomerp.components.bulk_update_objects import bulk_update_objects
from django.contrib.auth.models import AnonymousUser
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from bloomerp.components.bulk_upload_table import bulk_upload_table
from django.db import models
from django.apps import apps
from django.db import connection
from django.core.management import call_command
from django.core.files.uploadedfile import SimpleUploadedFile


class BookmarkComponentTests(TestCase):
    def setUp(self):
        self.url = reverse('components_bookmark')
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='testuser', password='12345')
        self.content_type = ContentType.objects.create(app_label='test_app', model='testmodel')

    def test_bookmark_creation(self):
        request = self.factory.post(self.url, {
            'content_type_id': self.content_type.id,
            'object_id': 1
        }, content_type='application/json')
        request.user = self.user
        response = bookmark(request)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(Bookmark.objects.filter(user=self.user, content_type=self.content_type, object_id=1).exists())

    def test_bookmark_deletion(self):
        Bookmark.objects.create(user=self.user, content_type=self.content_type, object_id=1)
        request = self.factory.post(self.url, {
            'content_type_id': self.content_type.id,
            'object_id': 1
        }, content_type='application/json')
        request.user = self.user
        response = bookmark(request)
        self.assertEqual(response.status_code, 200)
        self.assertFalse(Bookmark.objects.filter(user=self.user, content_type=self.content_type, object_id=1).exists())

    def test_bookmark_view_permission_denied(self):
        request = self.factory.get(self.url, {
            'content_type_id': self.content_type.id,
            'object_id': 1
        })
        request.user = self.user
        response = bookmark(request)
        self.assertEqual(response.status_code, 403)

    def test_bookmark_invalid_request(self):
        request = self.factory.post(self.url, {
            'content_type_id': 'invalid',
            'object_id': 'invalid'
        }, content_type='application/json')
        request.user = self.user
        response = bookmark(request)
        self.assertEqual(response.status_code, 400)

    def test_bookmark_unauthenticated(self):
        request = self.factory.post(self.url, {
            'content_type_id': self.content_type.id,
            'object_id': 1
        }, content_type='application/json')
        request.user = AnonymousUser()
        response = bookmark(request)
        self.assertEqual(response.status_code, 401)

# ----------------------------
# BULK UPDATE OBJECTS COMPONENT TESTS
# ----------------------------
class BulkUpdateObjectsTests(TestCase):
    def setUp(self):
        self.url = reverse('components_bulk_update_objects')
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='testuser', password='12345', is_superuser=True)
        self.content_type = ContentType.objects.get_for_model(File)
        self.file1 = File.objects.create(name='file1')
        self.file2 = File.objects.create(name='file2')
        self.message_middleware = MessageMiddleware(lambda request: None)
        self.session_middleware = SessionMiddleware(lambda request: None)

    def _add_middleware(self, request):
        self.session_middleware.process_request(request)
        request.session.save()
        self.message_middleware.process_request(request)

    def test_bulk_update_objects_success(self):
        request = self.factory.post(self.url+f'?content_type_id={self.content_type.id}&form_prefix=new_name', {
            'object_ids': [self.file1.id, self.file2.id],
        })
        request.user = self.user
        self._add_middleware(request)
        response = bulk_update_objects(request)
        self.assertEqual(response.status_code, 200)
        self.file1.refresh_from_db()
        self.file2.refresh_from_db()
        self.assertEqual(self.file1.name, 'new_name')
        self.assertEqual(self.file2.name, 'new_name')

    def test_bulk_delete_objects_success(self):
        request = self.factory.post(self.url+f'?content_type_id={self.content_type.id}&form_prefix=new_name', {
            'content_type_id': self.content_type.id,
            'object_ids': [self.file1.id, self.file2.id],
            'delete_objects': 'true'
        })
        request.user = self.user
        self._add_middleware(request)
        response = bulk_update_objects(request)
        self.assertEqual(response.status_code, 200)
        self.assertFalse(File.objects.filter(id__in=[self.file1.id, self.file2.id]).exists())

    def test_bulk_update_objects_permission_denied(self):
        self.user.user_permissions.clear()
        request = self.factory.post(self.url+f'?content_type_id={self.content_type.id}&form_prefix=new_name', {
            'content_type_id': self.content_type.id,
            'object_ids': [self.file1.id, self.file2.id],
            'form_prefix_name': 'new_name'
        })
        request.user = self.user
        self._add_middleware(request)
        response = bulk_update_objects(request)
        self.assertEqual(response.status_code, 403)

    def test_bulk_delete_objects_permission_denied(self):
        self.user.user_permissions.clear()
        request = self.factory.post(self.url+f'?content_type_id={self.content_type.id}&form_prefix=new_name', {
            'content_type_id': self.content_type.id,
            'object_ids': [self.file1.id, self.file2.id],
            'delete_objects': 'true'
        })
        request.user = self.user
        self._add_middleware(request)
        response = bulk_update_objects(request)
        self.assertEqual(response.status_code, 403)

    def test_bulk_update_objects_invalid_form_prefix(self):
        request = self.factory.post(self.url+f'?content_type_id={self.content_type.id}&form_prefix=new_name', {
            'content_type_id': self.content_type.id,
            'object_ids': [self.file1.id, self.file2.id]
        })
        request.user = self.user
        self._add_middleware(request)
        response = bulk_update_objects(request)
        self.assertEqual(response.status_code, 400)

    def test_bulk_update_objects_invalid_content_type(self):
        request = self.factory.post(self.url+f'?content_type_id=invalid&form_prefix=new_name', {
            'content_type_id': 'invalid',
            'object_ids': [self.file1.id, self.file2.id],
            'form_prefix_name': 'new_name'
        })
        request.user = self.user
        self._add_middleware(request)
        response = bulk_update_objects(request)
        self.assertEqual(response.status_code, 400)

    def test_bulk_update_objects_unauthenticated(self):
        request = self.factory.post(self.url+f'?content_type_id={self.content_type.id}&form_prefix=new_name', {
            'content_type_id': self.content_type.id,
            'object_ids': [self.file1.id, self.file2.id],
            'form_prefix_name': 'new_name'
        })
        request.user = AnonymousUser()
        self._add_middleware(request)
        response = bulk_update_objects(request)
        self.assertEqual(response.status_code, 401)

# ----------------------------
# BULK UPLOAD TABLE COMPONENT TESTS
# ----------------------------
class TestBulkUploadTableComponent(TestCase):
    # Create test model
    class A(models.Model):
        name = models.CharField(max_length=255)
        age = models.IntegerField()

        def __str__(self):
            return self.name
        
    # Create model that inherits from BloomerpModel
    class B(BloomerpModel):
        name = models.CharField(max_length=255)
        age = models.IntegerField()

        def __str__(self):
            return self.name

    def setUp(self):
        # Component URL
        self.url = reverse('components_bulk_upload_table')

        # Create users
        self.superuser = User.objects.create_user(username='testuser', password='12345', is_superuser=True)
        self.user = User.objects.create_user(username='testuser2', password='12345')
        
        # Create request factory
        self.factory = RequestFactory()

        # Get content type for model A and B
        self.content_type_a = ContentType.objects.get_for_model(self.A)
        self.content_type_b = ContentType.objects.get_for_model(self.B)

        # Save content types to database
        self.content_type_a.save()
        self.content_type_b.save()
        self.valid_file = SimpleUploadedFile("temp.csv", b'name,age\ntest,20\n', content_type="text/csv")
        self.invalid_file = SimpleUploadedFile("invalid.csv", b'invalid content', content_type="text/csv")


    def test_authorization_failed(self):
        '''Tests that the component only allows users with the correct permissions to access it'''
        # Create request with normal user
        request_1 = self.factory.post(
            self.url,
            data = {
                'bulk_upload_content_type_id': self.content_type_a.pk
            }
            )
        request_1.user = self.user

        # Test that normal user gets 403
        response_1 = bulk_upload_table(request_1)
        self.assertEqual(response_1.status_code, 403)

    def test_invalid_content_type(self):
        '''Tests that the component only allows valid content types'''
        # Create request with superuser
        request_2 = self.factory.post(
            self.url,
            data = {
                'bulk_upload_file': 'test',
                'bulk_upload_content_type_id': 'invalid'
            }
            )
        
        request_2.user = self.superuser

        # Test that superuser gets 400
        response_2 = bulk_upload_table(request_2)
        self.assertEqual(response_2.status_code, 400)

    def test_invalid_file(self):
        '''Tests that the component only allows valid files'''
        # Create request with superuser
        request = self.factory.post(
            self.url,
            data = {
                'bulk_upload_content_type_id': self.content_type_a.pk,
                'bulk_upload_file': SimpleUploadedFile("invalid.csv", b'invalid content', content_type="text/csv")
            },
            format='multipart'
        )
        request.user = self.superuser

        # Test that superuser
        response = bulk_upload_table(request)
        self.assertEqual(response.content.startswith(b'Error importing file:'), True)

    def test_valid_file(self):
        '''Tests that the component only allows valid files'''
        # Create request with superuser
        request = self.factory.post(
            self.url,
            data = {
                'bulk_upload_content_type_id': self.content_type_a.pk,
                'bulk_upload_file': self.valid_file
            },
            format='multipart'
        )
        request.user = self.superuser

        # Test that superuser
        response = bulk_upload_table(request)
        self.assertEqual(response.status_code, 200)

    def test_valid_form_submission(self):
        '''Tests that the component will create objects from a valid form submission'''
        # Create formset
        from django.forms.models import modelformset_factory

        data = [
            {'name': 'test1', 'age': 20},
            {'name': 'test2', 'age': 30},
            {'name': 'test3', 'age': 40}
        ]


        Formset = modelformset_factory(self.A, fields=['name', 'age'], extra=len(data))

        formset = Formset(queryset=self.A.objects.none(), initial=data, prefix = 'bulk_upload')

        # Create request with superuser
        request = self.factory.post(
            self.url,
            data = {
                'bulk_upload_content_type_id': self.content_type_a.pk,
                'bulk_upload_fields': ['name', 'age']
            },
        )
        request.user = self.superuser

        # Test that superuser
        response = bulk_upload_table(request)
        self.assertEqual(response.status_code, 200)

# ----------------------------
# SEARCH RESULTS COMPONENT TESTS
# ----------------------------
from bloomerp.components.search_results import search_results
class TestSearchResultsComponent(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='testuser', password='12345', is_superuser=True)
        self.content_type = ContentType.objects.create(app_label='test_app', model='testmodel')
        Link.objects.create(name='Test Link', url='test/url', level='LIST', content_type=self.content_type)
        self.url = reverse('components_search_results')


    def test_no_query_provided(self):
        request = self.factory.get(self.url)
        request.user = self.user
        response = search_results(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'No query provided', response.content)

    def test_list_level_links(self):
        request = self.factory.get(self.url, {'search_results_query': '/testmodel'})
        request.user = self.user
        response = search_results(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'Test Link', response.content)

    def test_app_level_links(self):
        Link.objects.create(name='App Link', url='app/url', level='APP')
        request = self.factory.get(self.url, {'search_results_query': '//A'})
        request.user = self.user
        response = search_results(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'App Link', response.content)

    def test_model_string_search(self):
        # Assuming your model has a string_search method
        request = self.factory.get(self.url, {'search_results_query': 'test'})
        request.user = self.user
        response = search_results(request)
        self.assertEqual(response.status_code, 200)
        # Add more assertions based on your model's string_search results


# Generated by Django 5.1.1 on 2024-09-17 20:53

import bloomerp.models.fields
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0014_alter_file_file"),
        ("test_app", "0007_alter_order_customer"),
    ]

    operations = [
        migrations.CreateModel(
            name="Accommodation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(blank=True, null=True)),
                ("price", models.DecimalField(decimal_places=2, max_digits=10)),
                (
                    "image",
                    bloomerp.models.fields.BloomerpFileField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="bloomerp.file",
                    ),
                ),
            ],
            options={
                "permissions": (
                    (
                        "bulk_delete%(class)s",
                        "Can delete multiple objects from %(class)s at once",
                    ),
                    (
                        "bulk_change%(class)s",
                        "Can update multiple objects from %(class)s at once",
                    ),
                ),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Employee",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("first_name", models.CharField(max_length=255)),
                ("last_name", models.CharField(max_length=255)),
                ("email", models.EmailField(max_length=254)),
                (
                    "identification_document",
                    bloomerp.models.BloomerpFileField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="bloomerp.file",
                    ),
                ),
            ],
            options={
                "permissions": (
                    (
                        "bulk_delete%(class)s",
                        "Can delete multiple objects from %(class)s at once",
                    ),
                    (
                        "bulk_change%(class)s",
                        "Can update multiple objects from %(class)s at once",
                    ),
                ),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="EmployeeOnAccommodation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("check_in", models.DateField()),
                ("check_out", models.DateField()),
                (
                    "accommodation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="test_app.accommodation",
                    ),
                ),
                (
                    "employee",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="test_app.employee",
                    ),
                ),
            ],
            options={
                "permissions": (
                    (
                        "bulk_delete%(class)s",
                        "Can delete multiple objects from %(class)s at once",
                    ),
                    (
                        "bulk_change%(class)s",
                        "Can update multiple objects from %(class)s at once",
                    ),
                ),
                "abstract": False,
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-12-29 21:05

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0059_documenttemplate_avatar_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="documenttemplate",
            name="save_to_folder",
            field=models.ForeignKey(
                blank=True,
                help_text="Signifies to which folder a file generated from the template needs to be saved upon creation.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="bloomerp.filefolder",
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2025-01-11 00:00

import django.db.models.deletion
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0067_alter_todo_required_by"),
    ]

    operations = [
        migrations.CreateModel(
            name="AIConversation",
            fields=[
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("datetime_updated", models.DateTimeField(auto_now=True)),
                ("title", models.CharField(default="AI Conversation", max_length=255)),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("conversation_history", models.JSONField()),
                (
                    "conversation_type",
                    models.CharField(
                        choices=[
                            ("sql", "SQL"),
                            ("document_template", "Document Template Generator"),
                            ("tiny_mce_content", "TinyMCE Content Generator"),
                            ("bloom_ai", "Bloom AI"),
                            ("code", "Code Generator"),
                        ],
                        default="bloom_ai",
                        max_length=20,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="%(class)s_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="%(class)s_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "AI conversation",
                "verbose_name_plural": "AI conversations",
                "db_table": "bloomerp_ai_conversation",
                "managed": True,
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-12-16 18:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0023_alter_customer_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="accommodation",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="customer",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="employee",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="employeeonaccommodation",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="equipment",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="function",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="order",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="product",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-20 22:07

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "test_app",
            "0019_alter_accommodation_options_alter_customer_options_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="accommodation",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                )
            },
        ),
        migrations.AlterModelOptions(
            name="customer",
            options={"verbose_name": "customer", "verbose_name_plural": "customers"},
        ),
        migrations.AlterModelOptions(
            name="employee",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                )
            },
        ),
        migrations.AlterModelOptions(
            name="employeeonaccommodation",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                )
            },
        ),
        migrations.AlterModelOptions(
            name="equipment",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                )
            },
        ),
        migrations.AlterModelOptions(
            name="function",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                )
            },
        ),
        migrations.AlterModelOptions(
            name="order",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                )
            },
        ),
        migrations.AlterModelOptions(
            name="product",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                )
            },
        ),
    ]



# ---------------------------------
# The URL patterns for the BloomerpEngine
# This file is responsible for generating the URL patterns for the BloomerpEngine
# ---------------------------------
from django.urls import include, path,  register_converter
from django.contrib.auth import views as auth_views
from django.contrib.contenttypes.models import ContentType
from bloomerp.components import datatable, todos
from bloomerp.components.document_templates import generate_document_template
from bloomerp.components.llm import llm_executor, ai_conversations
from bloomerp.views.document_templates import router as document_template_router
from django.db.models import Model
from bloomerp.utils.models import (
    get_attribute_name_for_foreign_key, 
    model_name_plural_underline, 
    model_name_plural_slug,
    get_foreign_occurences_for_model,
    get_model_dashboard_view_url,
    get_bulk_upload_view_url,
    get_base_model_route,
    get_detail_view_url,
    get_create_view_url,
    get_list_view_url,
    get_update_view_url
    )
from bloomerp.utils.api import generate_serializer, generate_model_viewset_class
from bloomerp.views.api import BloomerpModelViewSet
from bloomerp.models import Link, User
from bloomerp.utils.urls import IntOrUUIDConverter
from rest_framework.routers import DefaultRouter
from bloomerp.utils.router import _get_routers_from_settings, RouteFinder, BloomerpRouterHandler

# Register the custom URL converter
register_converter(IntOrUUIDConverter, 'int_or_uuid') # Register the custom URL converter
drf_router = DefaultRouter()

# Get the base URL from the settings
from django.conf import settings
BASE_URL = settings.BLOOMERP_SETTINGS.get('BASE_URL', '')

# Custom routers
from bloomerp.views.workspace import router as dashboard_router
from bloomerp.views.core import router as core_router
from bloomerp.views.widgets import router as widget_router
from bloomerp.views.auth import router as auth_router
from django.apps import apps
custom_routers = _get_routers_from_settings()


custom_routers.append(document_template_router)
custom_routers.append(dashboard_router)
custom_routers.append(core_router)
custom_routers.append(widget_router)
custom_routers.append(auth_router)


custom_router_handler = BloomerpRouterHandler(custom_routers)
custom_router_handler.generate_links()

# ---------------------------------
# START GENERATION OF URL PATTERNS
#
#           ___/\___
#          |        |
#         /|  O  O  |\      ankers away!
#        / |   ||   | \
#       /  |  ----  |  \
#      |   |  ||||  |   |
#      |  /|  ||||  |\  |
#      \/  |________|  \/
#           |      |
#           |______|
#           |  ||  |
#           |  ||  |
#          /|  ||  |\
#         / |______| \
#         |__________|
#
#           *   *   *
#          *  * * *  *
#         *  *  *  *  *
#        *  *   *   *  *
#       *  *    *    *  *
# ---------------------------------



# ---------------------------------
# Auth related URL patterns
# ---------------------------------
from django.conf import settings


from django.urls import reverse_lazy
urlpatterns = [
    path(settings.LOGIN_URL, auth_views.LoginView.as_view(
            template_name='auth_views/login_view.html',
            next_page=reverse_lazy('bloomerp_home_view')
            ), name='login'),
    path('logout/',auth_views.LogoutView.as_view(next_page=reverse_lazy('login')), name='logout'),
]

# ---------------------------------
# App level routes
# ---------------------------------
app_level_routes = custom_router_handler.get_app_routes()
for route in app_level_routes:
    urlpatterns.append(route.create_urlpattern())


# ---------------------------------
# Model related URL patterns
# ---------------------------------
content_types = ContentType.objects.all()


for content_type in content_types:
    if content_type.model_class():
        # Detail view patterns
        detail_view_patterns = []
        
        # Initialize some variables
        model : Model = content_type.model_class()
        base_model_route = get_base_model_route(model, include_slash=False) # Example: CustomerInvoice -> customer-invoices
        model_name_underline = model_name_plural_underline(model)
        model_name_plural = model._meta.verbose_name_plural
        model_name = model._meta.verbose_name

        
        # ---------------------------------
        # Custom routes
        # ---------------------------------
        custom_detail_view_routes = custom_router_handler.get_detail_view_routes_for_model(model)
        custom_list_view_routes = custom_router_handler.get_list_view_routes_for_model(model)
            
        for route in custom_detail_view_routes:
            # create the URL pattern
            urlpatterns.append(route.create_urlpattern())


        for route in custom_list_view_routes:
            # create the URL pattern
            urlpatterns.append(route.create_urlpattern())
            

        # ---------------------------------
        # Rest framework URL patterns
        # ---------------------------------
        # Generate the serializer class
        serializer_class = generate_serializer(model)

        # Generate the viewset class
        ApiViewSet = generate_model_viewset_class(
            model=model,
            serializer=serializer_class,
            base_viewset=BloomerpModelViewSet
        )

        try:
            drf_router.register(
                prefix = model_name_plural_underline(model), 
                viewset = ApiViewSet, 
                basename=model_name_plural_underline(model)
                )
        except:
            pass

        # ---------------------------------
        # Add the model to the admin dashboard
        # ---------------------------------
        from django.contrib import admin
        from django.contrib.admin.sites import AlreadyRegistered
        try:
            admin.site.register(model)
        except AlreadyRegistered:
            pass


# ---------------------------------
# API URL patterns
# ---------------------------------
urlpatterns += [
    path('api/', include(drf_router.urls))
]

# ---------------------------------
# Route decorator
# ---------------------------------
# Get current dir
import os
current_dir = os.path.dirname(os.path.abspath(__file__))
components_dir = os.path.join(current_dir, 'components')

route_finder = RouteFinder(directory=components_dir, url_route_prefix='components/', url_name_prefix='components')
urlpatterns += route_finder.generate_urlpatterns()



# ---------------------------------
# Create path
# ---------------------------------

BLOOMERP_URLPATTERNS = path(BASE_URL, include(urlpatterns))



# Generated by Django 5.1.1 on 2024-11-21 17:14

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Parent",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("age", models.IntegerField()),
                ("is_active", models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name="Child",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("age", models.IntegerField()),
                ("is_active", models.BooleanField(default=True)),
                # ...existing code...
                ("status", models.CharField(default="draft", max_length=50)),
                # ...existing code...
                ("date", models.DateField()),
                (
                    "parent",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="bloomerp_tests.parent",
                    ),
                ),
            ],
        ),
    ]



from cryptography.fernet import Fernet

class BloomerpEncryptionSuite:
    def __init__(self, key:str=None):
        '''Initializes the encryption suite'''
        self.key = key
        self.suite = Fernet(self.key)

    def encrypt_primary_key(self,pk):
        '''Encrypts a primary key integer, returns encrypted pk string'''
        encrypted = self.suite.encrypt(str(pk).encode())
        return encrypted.decode()

    def decrypt_primary_key(self,encrypted_pk):
        '''Decrypts an encrypted pk that is in a string format'''
        pk = self.suite.decrypt(encrypted_pk.encode())
        return int(pk.decode())


# Generated by Django 5.1.1 on 2024-12-08 22:49

import bloomerp.models.fields
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0052_alter_sqlquery_options"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="documenttemplate",
            name="standard_document",
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="footer",
            field=models.TextField(
                blank=True, help_text="Footer of the document template.", null=True
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="free_variables",
            field=models.ManyToManyField(
                blank=True,
                help_text="A free variable is a variable that is not from a model, and can be inserted in the template at creation time.",
                null=True,
                to="bloomerp.documenttemplatefreevariable",
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="model_variable",
            field=models.ForeignKey(
                default=1,
                help_text="Model variable of the document template.",
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.contenttype",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="name",
            field=models.CharField(
                help_text="Name of the document template.", max_length=100
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="styling",
            field=models.ForeignKey(
                blank=True,
                help_text="Styling of the document template.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="bloomerp.documenttemplatestyling",
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="template",
            field=bloomerp.models.fields.TextEditorField(
                default="Hello world",
                help_text="Content of the template, including the variables.",
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="template_header",
            field=models.ForeignKey(
                blank=True,
                help_text="Header of the document template.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="bloomerp.documenttemplateheader",
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplateheader",
            name="header",
            field=models.ImageField(
                help_text="Image of the header.", upload_to="document_templates/headers"
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplateheader",
            name="name",
            field=models.CharField(
                help_text="Name of the template header.", max_length=100
            ),
        ),
    ]



from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericRelation, GenericForeignKey
from django.db.models.query import QuerySet
from bloomerp.models import mixins
import os
import uuid
from django.contrib.auth.models import Permission


# ---------------------------------
# Bloomerp Model (abstract)
# ---------------------------------
class BloomerpModel(
    mixins.TimestampedModelMixin,
    mixins.StringSearchModelMixin,
    mixins.UserStampedModelMixin,
    mixins.AbsoluteUrlModelMixin,
    mixins.AvatarModelMixin,
    models.Model,
):
    '''
    Base model for all Bloomerp models.
    '''
    class Meta:
        abstract = True
        default_permissions = ('add', 'change', 'delete', 'view', 'bulk_change', 'bulk_delete', 'bulk_add', 'export')
    
    files = GenericRelation("bloomerp.File")
    comments = GenericRelation("bloomerp.Comment")


    form_layout : dict = None

    @classmethod
    def _validate_form_layout(cls) -> tuple[bool, list[str]]:
        """Validates the whether the form layout is correct."""
        if not cls.form_layout:
            return True, []

        EXCEPTION_FIELDS = ['created_by','updated_by','datetime_created','datetime_updated']

        fields = cls._meta.concrete_fields + cls._meta.many_to_many

        field_names = [field.name for field in fields]

        fields_in_form_layout = [] # List of fields that are contained in the form layout

        for field_section, field_list in cls.form_layout.items():
            for field in field_list:
                fields_in_form_layout.append(field)

        missing_fields = set(fields_in_form_layout) - set(field_names) - set(EXCEPTION_FIELDS)

        if missing_fields:
            return False, list(missing_fields)
        else:
            return True, []
        
    @classmethod
    def _get_form_layout(cls) -> dict:
        if cls._validate_form_layout()[0] == False:
            return False
        
        enhanced_layout = {}
        for title, field_names in cls.form_layout.items():
            required = False
            for field in field_names:
                if not cls._meta.get_field(field).null:
                    required = True
            
            enhanced_layout[title] = {
                "required" : required,
                "fields" : field_names
            }

        return enhanced_layout


# ---------------------------------
# ApplicationField Model
# ---------------------------------
class ApplicationField(models.Model):
    """
    ApplicationField is a model that stores information about fields and attributes in the Django model.
    """

    allow_string_search = False

    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = "bloomerp_application_field"

    FIELD_TYPES = [
        ("Property", "Property"),
        ("AutoField", "AutoField"),
        ("ForeignKey", "ForeignKey"),
        ("FileField", "FileField"),
        ("ImageField", "ImageField"),
        ("CharField", "CharField"),
        ("TextField", "TextField"),
        ("IntegerField", "IntegerField"),
        ("FloatField", "FloatField"),
        ("DecimalField", "DecimalField"),
        ("BooleanField", "BooleanField"),
        ("DateField", "DateField"),
        ("DateTimeField", "DateTimeField"),
        ("TimeField", "TimeField"),
        ("DurationField", "DurationField"),
        ("EmailField", "EmailField"),
        ("URLField", "URLField"),
        ("UUIDField", "UUIDField"),
        ("GenericIPAddressField", "GenericIPAddressField"),
        ("SlugField", "SlugField"),
        ("PositiveIntegerField", "PositiveIntegerField"),
        ("PositiveSmallIntegerField", "PositiveSmallIntegerField"),
        ("BigIntegerField", "BigIntegerField"),
        ("SmallIntegerField", "SmallIntegerField"),
        ("BinaryField", "BinaryField"),
        ("IPAddressField", "IPAddressField"),
        ("AutoField", "AutoField"),
        ("BigAutoField", "BigAutoField"),
        ("SmallAutoField", "SmallAutoField"),
        ("NullBooleanField", "NullBooleanField"),
        ("OneToOneField", "OneToOneField"),
        ("ManyToManyField", "ManyToManyField"),
        ("ArrayField", "ArrayField"),
        ("JSONField", "JSONField"),
        ("HStoreField", "HStoreField"),
        ("BinaryField", "BinaryField"),
        ("UUIDField", "UUIDField"),
        ("ForeignKey", "ForeignKey"),
        ("OneToOneField", "OneToOneField"),
        ("ManyToManyField", "ManyToManyField"),
        ("GenericRelation", "GenericRelation"),
        ("GenericForeignKey", "GenericForeignKey"),
        ("ForeignKey", "ForeignKey"),
        ("OneToOneField", "OneToOneField"),
        ("ManyToManyField", "ManyToManyField"),
        ("GenericRelation", "GenericRelation"),
        ("GenericForeignKey", "GenericForeignKey"),
        ("ForeignKey", "ForeignKey"),
        ("OneToOneField", "OneToOneField"),
        ("ManyToManyField", "ManyToManyField"),
        ("GenericRelation", "GenericRelation"),
        ("GenericForeignKey", "GenericForeignKey"),
    ]

    field = models.CharField(max_length=100)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    field_type = models.CharField(max_length=100, choices=FIELD_TYPES)
    related_model = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True, blank=True, related_name='related_models') # Related model for ForeignKey, OneToOneField, ManyToManyField
    meta = models.JSONField(null=True, blank=True)

    # Database related fields
    db_table = models.CharField(max_length=100, null=True, blank=True)
    db_field_type = models.CharField(max_length=100, null=True, blank=True)
    db_column = models.CharField(max_length=100, null=True, blank=True)

    def get_for_model(model:models.Model) -> QuerySet:
        """Returns application fields for a specific model"""
        return ApplicationField.objects.filter(
            content_type=ContentType.objects.get_for_model(model)
        )

    @property
    def title(self):
        return self.field.replace("_", " ").title()

    def __str__(self):
        return self.content_type.__str__() + " | " + str(self.field)

    @staticmethod
    def get_related_models(model: models.Model, skip_auto_created=True):
        """Returns all related models for a specific model"""
        content_type_id = ContentType.objects.get_for_model(model).pk
        qs = ApplicationField.objects.filter(
            meta__related_model=content_type_id
        ).exclude(content_type=content_type_id)
        if skip_auto_created:
            qs = qs.exclude(meta__auto_created=True)

        return qs

    @staticmethod
    def get_db_tables_and_columns(user= None) -> list[tuple[str, list[str]]]:
        """
        Returns a tuple for each database table.
        The tuple contains the table name and a tuple of the list of columns and there datatype.
        
        Args:
            user (User): The user object


        Example output:
        [
            ('auth_user', [('id', 'int'), ('username','varchar'), ...]),
            ('auth_group', [('id', 'int'), ('name','varchar'), ...]),
        ]

        """
        tables = []

        qs = ApplicationField.objects.filter(db_table__isnull=False)

        if user:
            content_types = user.get_content_types_for_user(permission_types=["view"])
            qs = qs.filter(content_type__in=content_types)


        for table in qs.values("db_table").distinct():
            table_name = table["db_table"]
            columns = ApplicationField.objects.filter(db_table=table_name).values_list(
                "db_column", "db_field_type"
            )
            tables.append((table_name, columns))
        return tables


# ---------------------------------
# File Model
# ---------------------------------
class File(
    mixins.TimestampedModelMixin, 
    mixins.StringSearchModelMixin,
    mixins.UserStampedModelMixin,
    models.Model,
):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = "bloomerp_file"

    search_fields = ['name']
    allow_string_search = True

    def upload_to(self, filename):
        '''Returns the upload path for the file'''
        # Can fetch this from settings in the future
        ROOT = 'bloomerp'

        if self.content_type is None:
            # Default folder for files with no content type
            folder = f'others'
        else:
            # Use the content type's app_label for organization
            folder = f'{self.content_type.app_label}'
        
        # Ensure unique file names
        unique_filename = f"{uuid.uuid4()}_{filename}"
        
        # Return the full path
        return f'{ROOT}/{folder}/{unique_filename}'
    
    # -----------------------------
    # File Fields
    # -----------------------------
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    file = models.FileField(upload_to=upload_to)
    name = models.CharField(max_length=100, null=True, blank=True)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True, blank=True)
    object_id = models.CharField(max_length=36, null=True, blank=True) # In order to support both UUID and integer primary keys
    content_object = GenericForeignKey("content_type", "object_id")
    persisted = models.BooleanField(default=False) # A field to indicate if the file is temporary or persisted

    # Created/updated utils
    meta = models.JSONField(blank=True, null=True)

    @property
    def url(self):
        return self.file.url

    @property
    def file_extension(self):
        """Returns the file extension of the file."""
        _, extension = os.path.splitext(self.file.name)
        return extension[1:]

    @property
    def size(self):
        """Returns the file size of the file."""
        try:
            return self.file.size
        except FileNotFoundError:
            return 0

    @property
    def size_str(self):
        """Returns the file size of the file in human readable format."""
        size = self.size
        if size < 1024:
            return f"{size} B"
        elif size < 1024 * 1024:
            return f"{size / 1024:.2f} KB"
        elif size < 1024 * 1024 * 1024:
            return f"{size / 1024 / 1024:.2f} MB"
        else:
            return f"{size / 1024 / 1024 / 1024:.2f} GB"

    def __str__(self):
        return str(self.name)


    def save(self, *args, **kwargs):
        # Check if a new file is being uploaded
        if self.pk:
            try:
                old_file = File.objects.get(pk=self.pk).file
                # If the file field is changed, delete the old file
                if old_file and old_file != self.file:
                    old_file.delete(save=False)
            except File.DoesNotExist:
                pass  # No old file exists

        # Set the name if not already set
        if not self.name:
            self.name = self.auto_name()
        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        # Delete the file when the object is deleted
        try:
            self.file.delete()
        except FileNotFoundError:
            pass
        super().delete(*args, **kwargs)

    def auto_name(self):
        """Returns the name of the file."""
        return self.file.name

    def get_accesible_files_for_user(
        query: str, 
        user, 
        folder=None, 
        content_type=None, 
        object_id=None
    ) -> QuerySet:
        """
        Returns a queryset of files that are accessible for the user.

        Args:
            query (str): The search query
            user (User): The user object
            folder (FileFolder): The folder object
            content_type (ContentType): The content type
            object_id (int): The object id

        Returns:
            QuerySet: A queryset of files
        """

        # Get the content types the user has access to
        content_types = user.get_content_types_for_user(permission_types=["view"])

        if folder:
            qs = folder.files.filter(content_type__in=content_types).order_by(
                "-datetime_created"
            )
        else:
            qs = File.objects.filter(content_type__in=content_types).order_by(
                "-datetime_created"
            )

        # Filter the queryset based on the content type
        if content_type:
            qs = qs.filter(content_type=content_type)

        # Filter the queryset based on the object id
        if object_id:
            qs = qs.filter(object_id=object_id)

        # Filter the queryset based on the query
        if query:
            qs = qs.filter(models.Q(name__icontains=query)).order_by(
                "-datetime_created"
            )

        return qs


# ---------------------------------
# Folder Model
# ---------------------------------
class FileFolder(
    mixins.TimestampedModelMixin,
    mixins.UserStampedModelMixin,
    mixins.AbsoluteUrlModelMixin,
    mixins.StringSearchModelMixin,
    models.Model,
):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = "bloomerp_file_folder"   

    name = models.CharField(max_length=255)
    files = models.ManyToManyField(File, related_name='folders', blank=True, null=True)
    parent = models.ForeignKey('self', on_delete=models.CASCADE, null=True, blank=True)
    content_types = models.ManyToManyField(ContentType, blank=True, null=True, help_text="Restrict folders to certain models.", verbose_name="Models")

    def __str__(self):
        return self.name
    
    string_search_fields = ['name']
    allow_string_search = True


    @property
    def parents(self):
        """Returns a list of parent folders."""
        parents = []
        parent = self.parent
        while parent:
            parents.append(parent)
            parent = parent.parent

        # Reverse the list to get the parents in the correct order
        return list(reversed(parents))
    
    @property
    def children(self):
        """Returns a list of child folders."""
        return FileFolder.objects.filter(parent=self)
    




# Generated by Django 5.1.1 on 2025-01-02 18:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0028_employee_gender"),
    ]

    operations = [
        migrations.AddField(
            model_name="order",
            name="date",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



from rest_framework import serializers
from django.db.models import Model
from rest_framework import viewsets
from bloomerp.utils.filters import dynamic_filterset_factory

def generate_serializer(model:Model):
    '''
    Dynamically generate a serializer class for a given model.
    '''

    # Dynamically create a Meta class
    meta_class = type('Meta', (object,), {
        'model': model,
        'fields': '__all__',
    })

    # Dynamically create the serializer class
    serializer_class = type(f'{model.__name__}Serializer', (serializers.ModelSerializer,), {
        'Meta': meta_class
    })

    return serializer_class


def generate_model_viewset_class(
        model:Model,
        serializer:serializers.ModelSerializer,
        base_viewset:viewsets.ModelViewSet
        ):
    '''
    Dynamically generate a viewset class for a given
    model.
    '''
    Class = type(f'{model.__name__}ViewSet', (base_viewset,), {
        'model': model,
        'serializer_class': serializer,
        'filterset_class': dynamic_filterset_factory(model)
    })



    return Class



# Generated by Django 5.1.1 on 2024-10-02 22:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0025_rename_folder_filefolder_alter_filefolder_options_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="filefolder",
            name="parent",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="bloomerp.filefolder",
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-26 17:54

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0011_function_employee_manager_employee_functions"),
    ]

    operations = [
        migrations.CreateModel(
            name="Equipment",
            fields=[
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("datetime_updated", models.DateTimeField(auto_now=True)),
                ("id", models.UUIDField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "permissions": (
                    (
                        "bulk_delete%(class)s",
                        "Can delete multiple objects from %(class)s at once",
                    ),
                    (
                        "bulk_change%(class)s",
                        "Can update multiple objects from %(class)s at once",
                    ),
                ),
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="employee",
            name="equipment",
            field=models.ManyToManyField(to="test_app.equipment"),
        ),
    ]



# Generated by Django 5.1.1 on 2024-12-13 18:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0055_documenttemplate_page_margin_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="documenttemplate",
            name="include_page_numbers",
            field=models.BooleanField(
                default=True,
                help_text="Signifies whether the page numbers are included or not.",
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="footer",
            field=models.TextField(
                default="", help_text="Footer content of the document template."
            ),
        ),
    ]



from bloomerp.models import BloomerpModel
from bloomerp.models.fields import StatusField
from django.db import models
import random

# Create your models here.

def generate_random_date():
    return f'{random.randint(1, 12)}/{random.randint(1, 28)}/{random.randint(2000, 2020)}'

class Parent(BloomerpModel):
    class Meta:
        db_table = 'bloomerp_tests_parent'

    name = models.CharField(max_length=255)
    age = models.IntegerField()
    is_active = models.BooleanField(default=True)
    status = StatusField(
        default='active',
        colored_choices=[
            ('active', 'Active', StatusField.GREEN),
            ('inactive', 'Inactive', StatusField.RED),
            ('pending', 'Pending', StatusField.YELLOW),
        ],
        max_length=255
    )
    date = models.DateField(default=generate_random_date)

class Many(BloomerpModel):
    class Meta:
        db_table = 'bloomerp_tests_many'

    name = models.CharField(max_length=255)
    age = models.IntegerField(default=random.randint(1, 100))
    is_active = models.BooleanField(default=True)
    date = models.DateField(default=generate_random_date)


    

class Child(BloomerpModel):
    class Meta:
        db_table = 'bloomerp_tests_child'

    name = models.CharField(max_length=255)
    age = models.IntegerField()
    is_active = models.BooleanField(default=True)
    date = models.DateField()
    parent_without_related_name = models.ForeignKey(Parent, on_delete=models.CASCADE)
    parent_with_related_name = models.ForeignKey(Parent, on_delete=models.CASCADE, related_name='children')
    many_to_many_without_related_name = models.ManyToManyField('Many')
    many_to_many_with_related_name = models.ManyToManyField('Many', related_name='children')
    status = StatusField(
        default='active',
        colored_choices=[
            ('active', 'Active', StatusField.GREEN),
            ('inactive', 'Inactive', StatusField.RED),
            ('pending', 'Pending', StatusField.YELLOW),
        ],
        max_length=255
    )
    




from django.test import TestCase
from bloomerp.models import User, Link, ContentType, BloomerpModel


# ----------------------------
# LINK MODEL TESTS
# ----------------------------
class LinkModelTests(TestCase):
    def setUp(self):
        pass

    def test_link_is_valid(self):
        # Create a valid link

        valid_link = Link.objects.filter(url="content_types_list").first()
        self.assertTrue(valid_link.is_valid())

        # Create an invalid link
        invalid_link = Link.objects.create(
            name="Invalid Link",
            url="invalid:url",
            level="APP",
            is_absolute_url=False,
            content_type=ContentType.objects.first()
        )
        self.assertFalse(invalid_link.is_valid())

    def test_invalid_workspace_link(self):
        # Create user
        user = User.objects.create_user(
            username="test_user",
            password="test_password"
        )

        # Create an invalid Link
        invalid_link = Link.objects.create(
            name="Invalid Link",
            url="content_types_list",
            level="APP",
            is_absolute_url=False,
            content_type=ContentType.objects.first()
        )

        # Get valid link
        valid_link = Link.objects.filter(url="content_types_list").first()

        # Create the workspace content
        content = {
            "content" : [
                {
                    "type": "link",
                    "size" : 12,
                    "data" :
                        {"link_id" : invalid_link.id}
                },
                {
                    "type": "link",
                    "size" : 12,
                    "data" :
                        {"link_id" : valid_link.id}
                },
                {
                    "type": "link_list",
                    "size" : 12,
                    "data" :
                        {"links" : [invalid_link.id, valid_link.id]}
                }
            ]
        }
        
        # Create the workspace
        workspace = Workspace.objects.create(
            name="Test Workspace",
            content=content,
            user = user
            )
        
        # Run the 
        workspace.remove_links_from_content(links=[invalid_link])

        # Check that the invalid link was removed
        removed_content = {
            "content" : [
                {
                    "type": "link",
                    "size" : 12,
                    "data" :
                        {"link_id" : valid_link.id}
                },
                {
                    "type": "link_list",
                    "size" : 12,
                    "data" :
                        {"links" : [valid_link.id]}
                }
            ]
        }

        self.assertEqual(workspace.content, removed_content)


from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.models import Comment
from django.contrib.contenttypes.models import ContentType

@route('comments')
def comments(request:HttpRequest) -> HttpResponse:
    # I know this is not the best way, but fuck it (chatGPT wrote that the fuck it part guys :)
    delete_comment_id = request.GET.get('delete_comment_id', None)
    edit_comment_id = request.GET.get('edit_comment_id', None)
    target = request.htmx.target
    

    # Get the variables
    if request.method == 'POST':
        # If comment should be deleted
        if delete_comment_id:
            comment = Comment.objects.get(pk=delete_comment_id)

            # Permission check
            if not request.user.has_perm('bloomerp.delete_comment') or comment.created_by!=request.user:
                return HttpResponse('Permission denied', status=403)
            
            content_type = comment.content_type
            object_id = comment.object_id
            comment.delete()

        elif edit_comment_id:
            comment = Comment.objects.get(pk=edit_comment_id)
            content = request.POST.get('content')

            # Permission check
            if not request.user.has_perm('bloomerp.change_comment') or comment.created_by!=request.user:
                return HttpResponse('Permission denied', status=403)

            content_type = comment.content_type
            object_id = comment.object_id

            comment.content = content
            comment.save()

        else:
            content = request.POST.get('content')
            object_id = request.POST.get('object_id')
            content_type_id = request.POST.get('content_type_id')

            # Get the content type
            content_type = ContentType.objects.get(id=content_type_id)

            # Create the comment
            user = request.user
            Comment.objects.create(
                content=content,
                created_by=user,
                updated_by=user,
                content_type=content_type,
                object_id=object_id
                )
            
        # Now filter for all comments for the object
        comments = Comment.objects.filter(content_type=content_type, object_id=object_id).order_by('-datetime_created')

        context = {
            'comments': comments,
            'target' : target
            }
        return render(request, 'components/comments.html', context)
    
    elif request.method == 'GET':
        object_id = request.GET.get('object_id')
        content_type_id = request.GET.get('content_type_id')

        # Get the content type
        content_type = ContentType.objects.get(id=content_type_id)

        # Permissions check
        # User needs to have permission to view comments and view content type
        if not request.user.has_perm('bloomerp.view_comment') or not request.user.has_perm(f'{content_type.app_label}.view_{content_type.model}'):
            return HttpResponse('Permission denied', status=403)


        # Now filter for all comments for the object
        comments = Comment.objects.filter(content_type=content_type, object_id=object_id).order_by('-datetime_created')

        context = {'comments': comments, 'target' : target}
        return render(request, 'components/comments.html', context)


import re
import uuid
from django.urls.resolvers import URLPattern

class IntOrUUIDConverter:
    regex = '[0-9]+|[0-9a-f-]{36}'

    def to_python(self, value):
        try:
            # Try to convert the value to an integer
            return int(value)
        except ValueError:
            # If it fails, try to convert it to a UUID
            return uuid.UUID(value)

    def to_url(self, value):
        return str(value)
    

def generate_detail_view_tabs(urlpatterns:list[URLPattern]):
    '''
    Generate DetailViewTab objects from urlpatterns.
    '''
    



from typing import Any
from django.forms import ValidationError
from bloomerp.models import ApplicationField, User, File, UserDetailViewTab, Link, UserListViewPreference
from django import forms
from django.contrib.contenttypes.models import ContentType
from django.forms.models import modelform_factory
from django.utils import timezone
from datetime import timedelta
from django.core.files import File as DjangoFile
from django.template.loader import get_template
from django.db.models import Model
from uuid import UUID
from bloomerp.utils.models import (
    get_file_fields_dict_for_model,
    get_bloomerp_file_fields_for_model,
    get_foreign_key_fields_for_model
    )
from django.contrib.postgres.fields import JSONField
from django.db.models import JSONField as DefaultJSONField
# ---------------------------------
# Bloomerp Bulk Upload Form
# ---------------------------------
class BulkUploadForm(forms.Form):

    def __init__(self, *args, **kwargs):
        super(BulkUploadForm, self).__init__(*args, **kwargs)
        for name, field in self.fields.items():
            if isinstance(field, forms.BooleanField):
                self.fields[name].widget = forms.Select(choices=[(True, 'True'), (False, 'False')])

        # Add delete all objects
        self.fields['delete_all'] = forms.BooleanField(required=False, label='Delete all selected objects',initial=False)

        # Remove last_updated_by field
        if 'last_updated_by' in self.fields:
            del self.fields['last_updated_by']

# ---------------------------------
# Object file form
# ---------------------------------

class ObjectFileForm(forms.ModelForm):
    class Meta:
        model = File
        fields = ['name', 'file']

    def __init__(self,
                 related_object:Model=None,
                 user:User=None,
                *args, **kwargs):

        self.related_object = related_object
        self.user = user

        super().__init__(*args, **kwargs)


    def save(self, commit=True):
        # Override the save method to set the content_type and object_id
        instance:File = super().save(commit=False)

        if self.related_object:
            instance.content_type = ContentType.objects.get_for_model(self.related_object)
            instance.object_id = self.related_object.pk
    
        if self.user:
            instance.uploaded_by = self.user


        if commit:
            instance.save()
        return instance

# ---------------------------------
# Bloomerp Model Form
# ---------------------------------
from bloomerp.widgets.foreign_key_widget import ForeignKeyWidget
from bloomerp.widgets.code_editor_widget import AceEditorWidget
from bloomerp.widgets.multiple_model_select_widget import MultipleModelSelect
from django.forms.widgets import DateInput, DateTimeInput
from bloomerp.forms.layouts import BloomerpModelformHelper

class BloomerpModelForm(forms.ModelForm):
    model:Model = None
    user:User = None
    instance:Model = None
    is_new_instance:bool = True

    def __init__(
            self, 
            model:Model, 
            user:User=None,
            apply_helper=True,
            hide_default_fields=True,
            *args, **kwargs):
        '''
        Args:
            model: The model for which the model form is made
            user: The user who is filling in the model form
            apply_helper: whether to apply the layout
            hide_default_fields: whether to hide the default fields (created_by, updated_by)
        
        '''


        # Set the model instance to the form instance
        self.model = model
        self._meta.model = model
        self.user = user

        super(BloomerpModelForm, self).__init__(*args, **kwargs)

        # Set the instance to the form instance
        instance:Model = kwargs.get('instance')
        if instance:
            self.instance = instance
            self.is_new_instance = False

        # Get all of the foreign key fields for the model
        self.foreign_key_fields = get_foreign_key_fields_for_model(self.model)


        # ---------------------------------
        # FOREIGN KEY FIELDS
        # ---------------------------------
        # Update the widgets for the foreign key fields
        for field in self.foreign_key_fields:
            # Get the related model
            if field.field in self.fields:
                related_model = field.meta['related_model']
                model = ContentType.objects.get(pk=related_model).model_class()
                self.fields[field.field].widget = ForeignKeyWidget(model=model)
        
        # ---------------------------------
        # MANY TO MANY FIELDS
        # ---------------------------------
        # Update the widgets for many to many fields
        for field in self._meta.model._meta.many_to_many:
            if field.name in self.fields:
                related_model = field.remote_field.model
                self.fields[field.name].widget = MultipleModelSelect(model=related_model)
    
        # ---------------------------------
        # FILE FIELDS
        # ---------------------------------
        # Update the widgets for the file fields
        self.file_fields = get_bloomerp_file_fields_for_model(self.model, output='list')

        # ---------------------------------
        # DATE AND DATETIME FIELDS
        # ---------------------------------
        for field_name, field in self.fields.items():
            if isinstance(field, forms.DateField):
                self.fields[field_name].widget = DateInput(attrs={'type': 'date'})
            elif isinstance(field, forms.DateTimeField):
                self.fields[field_name].widget = DateTimeInput(attrs={'type': 'datetime-local'})

        # ---------------------------------
        # JSON FIELDS
        # ---------------------------------
        # Update the widgets for the json fields
        for field_name, field in self.fields.items():
            # Check if the field is a JSONField
            model_field = self._meta.model._meta.get_field(field_name)
            
            if isinstance(model_field, (JSONField, DefaultJSONField)):
                # Apply the AceEditorWidget for JSON fields
                self.fields[field_name].widget = AceEditorWidget(language='json')

        # ---------------------------------
        # Hide created_by and updated_by fields
        # ---------------------------------
        if hide_default_fields:
            if 'created_by' in self.fields:
                del self.fields['created_by']
            if 'updated_by' in self.fields:
                del self.fields['updated_by']

        
        if self.model and apply_helper:
            helper = BloomerpModelformHelper(self.model)

            if helper.is_defined() and self.model:
                self.helper = helper

                
    def save(self, commit=True):
        '''
        Saves the form instance to the database

        '''

        instance = super(BloomerpModelForm, self).save(commit=False)

        # Check if the instance is new by checking if it has no primary key
        is_new_instance = instance.pk is None

        # Mark all temporary files as finalized after successful save
        def save_file_fields():
            if not instance.pk:
                raise ValueError("Instance must be saved before saving file fields")

            for field in self.file_fields:
                file:File = self.cleaned_data.get(field, None)
                if file:
                    # There is a new file so it has to be updated
                    file.persisted = True
                    file.content_type = ContentType.objects.get_for_model(self.model)
                    file.object_id = instance.pk
                    file.updated_by = self.user
                    file.created_by = self.user
                    file.save()
                else:
                    # There is no file, so we should delete the old file if it exists
                    old_file : File = getattr(instance, field, None)
                    if old_file:
                        old_file.delete()

        if commit:
            instance.save()
            save_file_fields()
        else:
            instance.save_file_fields = save_file_fields
            instance.is_new_instance = is_new_instance
        
        return instance

    def add_prefix(self, field_name):
        """
        Return the field name with a prefix appended.
        Overrides the default if the prefix contains "__".
        """
        if self.prefix and "__" in self.prefix:
            # Use "__" as the separator if the prefix contains "__"
            return f"{self.prefix}{field_name}"
        else:
            # Default behavior: use superclass method with hyphen separator
            return super().add_prefix(field_name)

# ---------------------------------
# Bloomerp Field Select Form
# ---------------------------------    
class BloomerpDownloadBulkUploadTemplateForm(forms.Form):
    file_type = forms.ChoiceField(choices=[('csv', 'CSV'), ('xlsx', 'Excel')])

    def __init__(self, model: Model = None, skip_fields=True, make_required=True, *args, **kwargs):
        super(BloomerpDownloadBulkUploadTemplateForm, self).__init__(*args, **kwargs)
        # Check if the content_type_id is provided in the request.POST
        content_type_id = kwargs.get('data', {}).get('content_type_id', None)
        if content_type_id:
            try:
                self.model = ContentType.objects.get(pk=content_type_id).model_class()
            except ContentType.DoesNotExist:
                self.model = None
        elif model:
            self.model = model
        else:
            raise ValueError("Model or content_type_id must be provided")

        if not self.model:
            raise ValueError("Model could not be determined.")

        # Init skip fields
        if skip_fields:
            skip_fields = [
                'created_by',
                'updated_by',
                'datetime_created',
                'datetime_updated',
            ]
        else:
            skip_fields = []

        # Model fields
        self.model_fields = []


        # Create model form and copy the fields
        ModelForm = modelform_factory(self.model, exclude=skip_fields)
        form_instance = ModelForm()

        # Add the fields to the form
        for field_name, field in form_instance.fields.items():
            if make_required:
                required = field.required
            else:
                required = False

            self.fields[field_name] = forms.BooleanField(
                label=field.label,
                required=field.required,
                initial=field.required,
                disabled=False,
                help_text=field.help_text
            )

            # Add the field to the model fields
            self.model_fields.append(field_name)

        # Add content type id as hidden field
        self.fields['content_type_id'] = forms.IntegerField(widget=forms.HiddenInput(), initial=ContentType.objects.get_for_model(self.model).pk)

    def get_selected_fields(self):
        # Run the clean method to get the selected fields
        if self.is_valid():
            # Only include fields that are checked and in the model fields
            return [field for field in self.model_fields if self.cleaned_data.get(field, False)]
        return []
    

    

# ---------------------------------
# Links select form
# ---------------------------------
class DetailLinksSelectForm(forms.Form):
    def __init__(self, content_type:ContentType, user:User, *args, **kwargs):
        super(DetailLinksSelectForm, self).__init__(*args, **kwargs)
        
        # Get all of the links that are available for the content type
        qs = Link.objects.filter(content_type=content_type, level='DETAIL') 
        
        for link in qs:
            if link.number_of_args() > 1:
                # Exclude links that require more than one argument
                qs = qs.exclude(pk=link.pk)
        
        # Get the links that the user has access to
        detail_view_links = UserDetailViewTab.get_detail_view_tabs(content_type=content_type, user=user).values_list('link_id', flat=True)

        self.fields['links'] = forms.ModelMultipleChoiceField(
            queryset=qs,
            widget=forms.CheckboxSelectMultiple
        )

        self.fields['links'].initial = detail_view_links

        # Add content type id as hidden field
        self.fields['content_type_id'] = forms.IntegerField(widget=forms.HiddenInput(), initial=content_type.pk)

        self.user = user


    def save(self) -> None:
        content_type_id = self.cleaned_data.get('content_type_id')
        content_type = ContentType.objects.get(pk=content_type_id)
        links = self.cleaned_data.get('links')

        # Get existing UserDetailViewTab objects for the user and content type
        existing_tabs = UserDetailViewTab.objects.filter(user=self.user, link__content_type=content_type)

        # Determine which links need to be added and which need to be removed
        existing_link_ids = set(existing_tabs.values_list('link_id', flat=True))
        selected_link_ids = set(links.values_list('id', flat=True))

        # Links to add
        links_to_add = selected_link_ids - existing_link_ids
        # Links to remove
        links_to_remove = existing_link_ids - selected_link_ids

        # Add new UserDetailViewTab objects
        for link_id in links_to_add:
            UserDetailViewTab.objects.create(user=self.user, link_id=link_id)

        # Remove UserDetailViewTab objects
        UserDetailViewTab.objects.filter(user=self.user, link_id__in=links_to_remove).delete()

        print("Links saved")


class ListViewFieldsSelectForm(forms.Form):
    def __init__(self, content_type:ContentType, user:User, *args, **kwargs):
        super(ListViewFieldsSelectForm, self).__init__(*args, **kwargs)
        
        # Get all of the Application fields that are available for the content type
        qs = ApplicationField.objects.filter(content_type=content_type)
        
        
        # Get the links that the user has access to
        application_fields = UserListViewPreference.objects.filter(user=user, application_field__in=qs).values_list('application_field_id', flat=True)
        
        self.fields['fields'] = forms.ModelMultipleChoiceField(
            queryset=qs,
            widget=forms.CheckboxSelectMultiple
        )

        self.fields['fields'].initial = application_fields

        # Add content type id as hidden field
        self.fields['content_type_id'] = forms.IntegerField(widget=forms.HiddenInput(), initial=content_type.pk)

        self.user = user


    def save(self) -> None:
        content_type_id = self.cleaned_data.get('content_type_id')
        content_type = ContentType.objects.get(pk=content_type_id)
        fields = self.cleaned_data.get('fields')

        # Get existing UserListViewPreference objects for the user and content type
        existing_preferences = UserListViewPreference.objects.filter(user=self.user, application_field__content_type=content_type)

        # Determine which fields need to be added and which need to be removed
        existing_field_ids = set(existing_preferences.values_list('application_field_id', flat=True))
        selected_field_ids = set(fields.values_list('id', flat=True))

        # Fields to add
        fields_to_add = selected_field_ids - existing_field_ids
        # Fields to remove
        fields_to_remove = existing_field_ids - selected_field_ids

        # Add new UserListViewPreference objects
        for field_id in fields_to_add:
            UserListViewPreference.objects.create(user=self.user, application_field_id=field_id)

        # Remove UserListViewPreference objects
        UserListViewPreference.objects.filter(user=self.user, application_field_id__in=fields_to_remove).delete()




# Generated by Django 5.1.1 on 2024-11-04 20:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0040_alter_userdetailviewtab_unique_together_and_more"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.AddField(
            model_name="filefolder",
            name="content_types",
            field=models.ManyToManyField(
                blank=True,
                help_text="Restrict folders to certain models.",
                null=True,
                to="contenttypes.contenttype",
                verbose_name="Models",
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-05 19:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0042_rename_kpi_widget_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="applicationfield",
            name="db_column",
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render, redirect
from django.http import HttpResponse, HttpRequest
from django.views.decorators.http import require_POST
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from bloomerp.models import File
from bloomerp.utils.document_templates import DocumentController
import base64
from django.utils.safestring import mark_safe


@login_required
@require_POST
@route('sign_file')
def sign_file(request:HttpRequest) -> HttpResponse:
    '''
    Post data to sign a file.
    
    POST Args:
        file_id (int): The id of the file to sign.
        signature (bytes): The bytes of the signature.
    '''
    # Some permissions check
    if not request.user.has_perm('bloomerp.view_file'):
        return HttpResponse('User does not have permission to view files')
    
    signature_data = request.POST.get('signature')

    try:
        file_id = request.POST.get('file_id')
    except:
        return HttpResponse('Invalid file id', status=200)

    try:
        # Get the bytes of the signature
        signature_data = base64.b64decode(signature_data.split(",")[1])

        # Get the file or return a 404
        file = File.objects.get(id=file_id)

        # Sign the file
        document_controller = DocumentController(user=request.user)

        # Sign the file
        signed_file = document_controller.sign_pdf(file, signature_data)    

        # Display success message
        messages.success(request, f'File has been signed')

        # Redirect user to page where request was made
        return redirect(request.META.get('HTTP_REFERER', '/'))
    
    except Exception as e:
        messages.error(request, f'Error signing file: {e}')
        return redirect(request.META.get('HTTP_REFERER', '/'))


# Generated by Django 5.1.1 on 2024-09-23 21:45

import bloomerp.models.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0018_documenttemplatestyling_documenttemplate_styling"),
    ]

    operations = [
        migrations.AlterField(
            model_name="documenttemplatestyling",
            name="styling",
            field=bloomerp.models.fields.CodeField(default="", language="css"),
        ),
    ]




from django.test import TestCase, RequestFactory
from django.contrib.auth.models import User
from bloomerp.views.core import BloomerpListView, BloomerpDetailOverviewView

class BloomerpViewsTests(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='testuser', password='12345')

    def test_bloomerp_list_view(self):
        request = self.factory.get('/list')
        request.user = self.user
        response = BloomerpListView.as_view()(request)
        self.assertEqual(response.status_code, 200)

    def test_bloomerp_detail_overview_view(self):
        request = self.factory.get('/overview')
        request.user = self.user
        response = BloomerpDetailOverviewView.as_view()(request)
        self.assertEqual(response.status_code, 200)


# Generated by Django 5.1.1 on 2024-10-30 18:10

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0037_alter_documenttemplatestyling_styling_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="UserDetailViewTabs",
        ),
    ]



from django.apps import AppConfig

class BloomerpConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "bloomerp"

    def ready(self) -> None:
        
        # import bloomerp.signals
        from bloomerp.utils.config import BloomerpConfigChecker
        checker = BloomerpConfigChecker()
        checker.check()
        
        





import pandas as pd
from django.core.files import File
from django.db import models
from io import BytesIO, StringIO, TextIOWrapper
import csv
import openpyxl
from django.core.exceptions import ValidationError
from bloomerp.models import User

class BloomerpModelIO:
    def __init__(self, model: models.Model):
        self.model = model

    def get_model_fields(self) -> list[models.Field]:
        """
        Get all of the fields for the that are editable and not an AutoField.
        """
        fields = self.model._meta.get_fields()

        # Filter out fields that are not editable and id fields
        return [field for field in fields if field.editable and not isinstance(field, models.AutoField)]

    def get_model_fields_as_str(self) -> list[str]:
        """
        Get all of the fields for the model as a list of strings.
        """
        return [field.name for field in self.get_model_fields()]

    def get_model_field_data_types(self) -> dict:
        """
        Get the field names and their corresponding data types for the model.
        """
        return {field.name: field for field in self.get_model_fields()}

    def export_to_csv(self, queryset=None, fields='__all__') -> bytes:
        """
        Export the model data to a CSV file in bytes format.
        If the queryset is None, export all of the model data.
        """
        if queryset is None:
            queryset = self.model.objects.all()

        if fields == '__all__':
            fields = [field.name for field in self.get_model_fields()]

        buffer = StringIO()
        writer = csv.writer(buffer)
        writer.writerow(fields)

        for obj in queryset:
            row = [getattr(obj, field) for field in fields]
            writer.writerow(row)

        csv_content = buffer.getvalue()
        buffer.close()

        return csv_content.encode('utf-8')  # Convert string to bytes

    def export_to_excel(self, queryset=None, fields='__all__') -> bytes:
        """
        Export the model data to an Excel file in bytes format.
        If the queryset is None, export all of the model data.
        """
        if queryset is None:
            queryset = self.model.objects.all()

        if fields == '__all__':
            fields = [field.name for field in self.get_model_fields()]

        buffer = BytesIO()
        workbook = openpyxl.Workbook()
        sheet = workbook.active
        sheet.append(fields)

        for obj in queryset:
            row = [getattr(obj, field).__str__() for field in fields]
            sheet.append(row)

        workbook.save(buffer)
        buffer.seek(0)
        return buffer.getvalue()

    def validate_row_data(self, row: dict, fields: dict) -> bool:
        """
        Validate a single row of data to check if it matches the field types of the model.
        """
        for field_name, field in fields.items():
            value = row.get(field_name, None)

            if value is None:
                if not field.null:
                    return False  # Non-nullable field is missing
            else:
                try:
                    # Perform type validation based on Django field type
                    if isinstance(field, (models.IntegerField, models.AutoField)):
                        int(value)  # Try to cast to int
                    elif isinstance(field, models.FloatField):
                        float(value)  # Try to cast to float
                    elif isinstance(field, models.BooleanField):
                        if not isinstance(value, bool):
                            return False  # Value must be boolean
                    elif isinstance(field, models.DateField):
                        pd.to_datetime(value, errors='raise')  # Validate date
                    elif isinstance(field, models.CharField):
                        str(value)  # Ensure it is string-like
                except (ValueError, TypeError):
                    return False  # Value does not match expected type

        return True

    def import_from_template(self, file: File) -> list[dict]:
        '''
        Imports data from a CSV or Excel file template and returns a list of dictionaries with data for objects.
        '''
        # Load the file into a pandas dataframe
        if file.name.endswith('.csv'):
            dataframe = pd.read_csv(TextIOWrapper(file.file, encoding='utf-8'))
        elif file.name.endswith('.xlsx'):
            dataframe = pd.read_excel(file)
        else:
            raise ValueError("Unsupported file type. Only CSV and Excel files are allowed.")
        
        # Give the dataframe to the template_is_valid method to check if the template is valid
        if not self._template_is_valid(dataframe):
            raise ValidationError("Invalid template. Please ensure all required fields are present.")
        

        # Fill NaN values with empty strings
        dataframe.fillna('', inplace=True)

        # Create a list of unsaved model instances
        instances = []
        for _, row in dataframe.iterrows():
            obj_data = {}
            m2m_fields = []
            for field in self.get_model_fields():
                field_name = field.name
                if field_name in row:
                    # Check if the field is a ManyToManyField, if so split the values by | and store them in a list
                    if isinstance(field, models.ManyToManyField):
                        if type(row[field_name]) == str:
                            obj_data[field_name] = row[field_name].split('|')
                        else:
                            # Only one value, so store it in a list
                            obj_data[field_name] = [row[field_name]]
                    else:
                        obj_data[field_name] = row[field_name]

            instances.append(obj_data)

        return instances

    def _template_is_valid(self, dataframe:pd.DataFrame) -> bool:
        # Get model fields as a list of strings, assuming they are comma-separated
        fields = self.get_model_fields_as_str()

        # Trim whitespace and convert columns to lowercase for consistent comparison
        df_columns = dataframe.columns.str.strip().str.lower()
        fields = [field.lower() for field in fields]

        print("File columns:", df_columns.tolist())
        print("Model fields:", fields)

        # Check if all of the model fields are present in the template (file columns)
        if not set(df_columns).issubset(fields):
            return False

        return True

    def create_model_template(self, file_type: str, fields='__all__') -> bytes:
        """
        Create a template for the model in the specified file format (CSV or Excel).

        Args:
            file_type (str): The file format to create the template in ('csv' or 'xlsx').
            fields (list): The fields to include in the template. If '__all__', include all fields.
        """
        # Validate fields
        model_fields = [field.name for field in self.get_model_fields()]
        if fields == '__all__':
            fields = model_fields
        else:
            # Ensure all provided fields are valid model fields
            invalid_fields = [field for field in fields if field not in model_fields]
            if invalid_fields:
                raise ValueError(f"Invalid fields: {', '.join(invalid_fields)}")

        if file_type == 'csv':
            buffer = StringIO()
            writer = csv.writer(buffer)
            writer.writerow(fields)
            csv_content = buffer.getvalue()
            buffer.close()
            return csv_content.encode('utf-8')  # Convert string to bytes

        elif file_type == 'xlsx':
            buffer = BytesIO()
            workbook = openpyxl.Workbook()
            sheet = workbook.active
            sheet.append(fields)
            workbook.save(buffer)
            buffer.seek(0)
            return buffer.getvalue()

        else:
            raise ValueError("Invalid file type. Must be either 'csv' or 'xlsx'.")
        


# Generated by Django 5.1.1 on 2024-09-27 11:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0022_alter_sqlquery_query"),
    ]

    operations = [
        migrations.AddField(
            model_name="userdashboardview",
            name="layout",
            field=models.JSONField(
                blank=True, help_text="The layout of the dashboard", null=True
            ),
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.models import ApplicationField
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType

@login_required
@route('list_view_filter')
def list_view_filter(request:HttpRequest) -> HttpResponse:
    '''
    Renders list view filter component.

    Variables:
        - content_type_id: The content type id of the model to be rendered in the data table
        - target : The target of the data table
    '''
    content_type_id = request.GET.get('content_type_id', None)
    target = request.GET.get('target', None)
    include_actions = request.GET.get('data_table_include_actions', False) # Actions will be rendered to the right of the table for each row
    user = request.user

    if not content_type_id:
        return HttpResponse("Content type id is required", status=400)
    if not target:
        return HttpResponse("Target is required", status=400)
    else:
        target = "#" + target

    # Some logic for permissions here based on content_type_id
    model = ContentType.objects.get(id=content_type_id).model_class()
    if not user.has_perm(f'{model._meta.app_label}.view_{model._meta.model_name}'):
        return HttpResponse('User does not have permission to view this data table')


    # Get the application fields for the content type
    application_fields = ApplicationField.objects.filter(content_type_id=content_type_id).exclude(field_type='Property')
    

    context = {
        'application_fields': application_fields,
        'target': target,
        'content_type_id': content_type_id,
        'include_actions': include_actions
    }
    return render(request, 'snippets/list_view_filter.html', context)


from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.utils.document_templates import DocumentController
from bloomerp.models import DocumentTemplate, File
from django.shortcuts import get_object_or_404
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_POST
from bloomerp.forms.document_templates import GenerateDocumentForm
import base64
from bloomerp.utils.requests import parse_bool_parameter

@login_required
@route('generate_document_template')
def generate_document_template(request:HttpRequest) -> HttpResponse:
    '''
    Component that generates a document template.

    GET Arguments:
        - document_template_id : The document template ID
        - add_persist_field : Whether to add the option to persist the document
    
    POST Arguments:
        - form : The form data
        - preview_content : Content to preview for the document template (optional)
        - add_persist_field : add_persist_field (optional)
        

    '''
    # Some permissions check
    if not request.user.has_perm('bloomerp.view_documenttemplate'):
        return HttpResponse('User does not have permission to view document templates')

    if request.method == 'POST':
        # Get document template from the form
        document_template_id = request.POST.get('document_template_id')
        document_template = get_object_or_404(DocumentTemplate, id=document_template_id)

        # Check whether there is preview content
        preview_content = request.POST.get('preview_content', None)
        
        # Get add persist field for later form creation
        add_persist_field = parse_bool_parameter(request.POST.get('add_persist_field'), True)

        if preview_content:
            document_template.template = preview_content

        # Get the form data
        form = GenerateDocumentForm(
            document_template=document_template, 
            data = request.POST, 
            files = request.FILES,
            add_persist_field=add_persist_field
            )
        
        # If form is not valid return from
        if not form.is_valid():
            return render(request, 'components/generate_document_template.html', context={'form': form})
        else:
            try:
                # Get the cleaned data from the form
                data = form.cleaned_data

                # Initialize the document controller
                document_controller = DocumentController(document_template, request.user)

                # Generate the document without persisting
                file: File = document_controller.create_document(
                    document_template=document_template,
                    free_variables = data, 
                    instance=form.instance,
                    persist=form.persist
                )

                # Ensure the file is associated
                if not file.file:
                    return HttpResponse('File not associated correctly', status=500)

                # Get the bytes of the file
                file_bytes = file.file.read()
                file_base64 = base64.b64encode(file_bytes).decode('utf-8')

                return render(request, 'components/generate_document_template.html', context={'file_bytes': file_base64, 'form':form})
            except TypeError as e:
                if 'RelatedManager' in str(e):
                    form.add_error(None, "Please make sure that when using a for-loop, '.all' is used at the end.")
                else:
                    form.add_error(None, str(e))
                return render(request, 'components/generate_document_template.html', context={'form': form})
            except Exception as e:
                form.add_error(None, str(e))
                return render(request, 'components/generate_document_template.html', context={'form': form})
            
    else:
        # Get the document template from the GET request
        document_template_id = request.GET.get('document_template_id')
        
        # Get the persist option
        add_persist_field = parse_bool_parameter(request.GET.get('add_persist_field'), False)

        # Get the document template
        document_template = get_object_or_404(DocumentTemplate, id=document_template_id)

        # Create the form
        form = GenerateDocumentForm(document_template=document_template, add_persist_field=add_persist_field)

        return render(request, 'components/generate_document_template.html', context={'form': form})



# Generated by Django 5.1.1 on 2025-01-10 19:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0066_rename_requested_for_todo_assigned_to"),
    ]

    operations = [
        migrations.AlterField(
            model_name="todo",
            name="required_by",
            field=models.DateField(
                blank=True,
                help_text="The date by which the todo is required",
                null=True,
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-17 20:19

import bloomerp.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0013_applicationfield_related_model_alter_file_file"),
    ]

    operations = [
        migrations.AlterField(
            model_name="file",
            name="file",
            field=models.FileField(upload_to=bloomerp.models.File.upload_to),
        ),
    ]



# Generated by Django 5.1.1 on 2024-10-06 13:27

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0013_alter_customer_name_alter_employee_first_name_and_more"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name="accommodation",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="accommodation",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="employee",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="employee",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="employeeonaccommodation",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="employeeonaccommodation",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="equipment",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="equipment",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="function",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="function",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="employee",
            name="manager",
            field=models.ForeignKey(
                blank=True,
                help_text="The manager of the employee",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="test_app.employee",
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2025-01-08 00:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0064_remove_todo_avatar"),
    ]

    operations = [
        migrations.AlterField(
            model_name="todo",
            name="priority",
            field=models.IntegerField(
                choices=[(1, "Low"), (2, "Medium"), (3, "High")],
                default=2,
                help_text="The priority of the todo",
            ),
        ),
        migrations.AlterField(
            model_name="todo",
            name="required_by",
            field=models.DateTimeField(
                blank=True,
                help_text="The date by which the todo is required",
                null=True,
            ),
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import BloomerpModel
from bloomerp.utils.models import string_search
from django.contrib.auth.decorators import login_required

@login_required
@route('fk_search_results')
def fk_search_results(request:HttpRequest) -> HttpResponse:
    '''
    Component that returns search results for a given query.
    This component is used for foreign key or many to many fields.
    '''

    Model : BloomerpModel = ContentType.objects.get_for_id(request.GET.get('content_type_id')).model_class()
    query = request.GET.get('fk_search_results_query')
    field_name = request.GET.get('field_name')
    search_type = request.GET.get('search_type','fk')

    if search_type not in ['fk','m2m']:
        pass


    if query:
        # Check if the model has a string_search method, it is inherited from BloomerpModel
        if not hasattr(Model, 'string_search'):
            # Add the string_search method to the model
            Model.string_search = classmethod(string_search)
        
        context = {
            'objects': Model.string_search(query)[:5]
        }
    else:
        context = {
            'objects': Model.objects.all()[:5]
        }

    context['field_name'] = field_name
    context['type'] = search_type

    return render(request, 'components/fk_search_results.html', context)


from django.forms.models import modelform_factory
from django.shortcuts import render
from bloomerp.utils.filters import dynamic_filterset_factory
from bloomerp.models import Workspace


def test_workspace(request):
    workspace = Workspace.objects.first()


    return render(request, 'test_view.html', {'workspace': workspace})




from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from django.views.decorators.http import require_POST
from django.contrib.auth.decorators import login_required
from bloomerp.models import Workspace

@login_required
@require_POST
@route('reset_workspace')
def reset_workspace(request:HttpRequest) -> HttpResponse:
    '''Component that resets a workspace to it's default state.
    
    POST Args:
        workspace_id (int): The id of the workspace to reset.
    '''
    # Permissions check
    if not request.user.has_perm('bloomerp.change_workspace'):
        return HttpResponse('User does not have permission to view work', status=200)

    print(request.POST)

    workspace_id = request.POST.get('workspace_id')

    print(workspace_id)

    try:
        workspace = Workspace.objects.get(id=workspace_id)
    except:
        return HttpResponse('Invalid workspace id', status=200)
    
    # Reset the workspace
    if workspace.content_type:
        workspace.content = Workspace.create_default_content_type_workspace(
            user=request.user,
            content_type=workspace.content_type,
            commit=False
        ).content
    else:
        workspace.content = Workspace.create_default_workspace(
            user=request.user,
            commit=False
        ).content

    workspace.save()
 


# Generated by Django 5.1.1 on 2024-09-15 20:25

import bloomerp.models
import django.contrib.auth.models
import django.contrib.auth.validators
import django.db.models.deletion
import django.utils.timezone
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={
                            "unique": "A user with that username already exists."
                        },
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        unique=True,
                        validators=[
                            django.contrib.auth.validators.UnicodeUsernameValidator()
                        ],
                        verbose_name="username",
                    ),
                ),
                (
                    "first_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="first name"
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="last name"
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        blank=True, max_length=254, verbose_name="email address"
                    ),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                ("avatar", models.ImageField(null=True, upload_to="users/")),
                (
                    "file_view_preference",
                    models.CharField(
                        choices=[("card", "Card View"), ("list", "List View")],
                        default="card",
                        max_length=20,
                    ),
                ),
                (
                    "date_view_preference",
                    models.CharField(
                        choices=[
                            ("d-m-Y", "Day-Month-Year (15-08-2000)"),
                            ("m-d-Y", "Month-Day-Year (08-15-2000)"),
                            ("Y-m-d", "Year-Month-Day (2000-08-15)"),
                        ],
                        default="d-m-Y",
                        max_length=20,
                    ),
                ),
                (
                    "datetime_view_preference",
                    models.CharField(
                        choices=[
                            (
                                "d-m-Y H:i",
                                "Day-Month-Year Hour:Minute (15-08-2000 12:30)",
                            ),
                            (
                                "m-d-Y H:i",
                                "Month-Day-Year Hour:Minute (08-15-2000 12:30)",
                            ),
                            (
                                "Y-m-d H:i",
                                "Year-Month-Day Hour:Minute (2000-08-15 12:30)",
                            ),
                        ],
                        default="d-m-Y H:i",
                        max_length=20,
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "db_table": "auth_user",
            },
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name="ApplicationField",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("field", models.CharField(max_length=100)),
                (
                    "field_type",
                    models.CharField(
                        choices=[
                            ("Property", "Property"),
                            ("AutoField", "AutoField"),
                            ("ForeignKey", "ForeignKey"),
                            ("FileField", "FileField"),
                            ("ImageField", "ImageField"),
                            ("CharField", "CharField"),
                            ("TextField", "TextField"),
                            ("IntegerField", "IntegerField"),
                            ("FloatField", "FloatField"),
                            ("DecimalField", "DecimalField"),
                            ("BooleanField", "BooleanField"),
                            ("DateField", "DateField"),
                            ("DateTimeField", "DateTimeField"),
                            ("TimeField", "TimeField"),
                            ("DurationField", "DurationField"),
                            ("EmailField", "EmailField"),
                            ("URLField", "URLField"),
                            ("UUIDField", "UUIDField"),
                            ("GenericIPAddressField", "GenericIPAddressField"),
                            ("SlugField", "SlugField"),
                            ("PositiveIntegerField", "PositiveIntegerField"),
                            ("PositiveSmallIntegerField", "PositiveSmallIntegerField"),
                            ("BigIntegerField", "BigIntegerField"),
                            ("SmallIntegerField", "SmallIntegerField"),
                            ("BinaryField", "BinaryField"),
                            ("IPAddressField", "IPAddressField"),
                            ("AutoField", "AutoField"),
                            ("BigAutoField", "BigAutoField"),
                            ("SmallAutoField", "SmallAutoField"),
                            ("NullBooleanField", "NullBooleanField"),
                            ("OneToOneField", "OneToOneField"),
                            ("ManyToManyField", "ManyToManyField"),
                            ("ArrayField", "ArrayField"),
                            ("JSONField", "JSONField"),
                            ("HStoreField", "HStoreField"),
                            ("BinaryField", "BinaryField"),
                            ("UUIDField", "UUIDField"),
                            ("ForeignKey", "ForeignKey"),
                            ("OneToOneField", "OneToOneField"),
                            ("ManyToManyField", "ManyToManyField"),
                            ("GenericRelation", "GenericRelation"),
                            ("GenericForeignKey", "GenericForeignKey"),
                            ("ForeignKey", "ForeignKey"),
                            ("OneToOneField", "OneToOneField"),
                            ("ManyToManyField", "ManyToManyField"),
                            ("GenericRelation", "GenericRelation"),
                            ("GenericForeignKey", "GenericForeignKey"),
                            ("ForeignKey", "ForeignKey"),
                            ("OneToOneField", "OneToOneField"),
                            ("ManyToManyField", "ManyToManyField"),
                            ("GenericRelation", "GenericRelation"),
                            ("GenericForeignKey", "GenericForeignKey"),
                        ],
                        max_length=100,
                    ),
                ),
                ("meta", models.JSONField(blank=True, null=True)),
                ("db_table", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "db_field_type",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_application_field",
                "managed": True,
            },
        ),
        migrations.CreateModel(
            name="File",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "file",
                    models.FileField(upload_to=bloomerp.models.File.upload_to),
                ),
                ("name", models.CharField(blank=True, max_length=100, null=True)),
                ("object_id", models.PositiveIntegerField(blank=True, null=True)),
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("meta", models.JSONField(blank=True, null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "uploaded_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_file",
                "managed": True,
            },
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.models import Bookmark, User
from django.views.decorators.csrf import csrf_exempt
from django.contrib.auth.decorators import login_required

@csrf_exempt
@login_required
@route('bookmark')
def bookmark(request:HttpRequest) -> HttpResponse:
    # Get variables
    user:User = request.user
    size = request.GET.get('size', 30)

    # Get the content_type_id and object_id from the request
    try:
        content_type_id = int(request.GET.get('content_type_id'))
        object_id = int(request.GET.get('object_id'))
    except:
        return HttpResponse(status=400)

    if not user.is_authenticated:
        return HttpResponse(status=401)
    
    # Check if the content_type_id and object_id are provided
    if not content_type_id or not object_id:
        return HttpResponse(status=400)


    if request.method == 'POST':
        if not request.user.has_perm('bloomerp.add_bookmark'):
            return HttpResponse(status=403)

        # Check if the bookmark allready exists
        # If it does, delete it
        # If it does not, create it
        bookmark = Bookmark.objects.filter(user=user, content_type_id=content_type_id, object_id=object_id)
        if bookmark.exists():
            bookmark.delete()
            bookmarked = False
        else:
            Bookmark.objects.create(user=user, content_type_id=content_type_id, object_id=object_id)
            bookmarked = True
    else:
        # Permissions check
        if not user.has_perm('bloomerp.view_bookmark'):
            return HttpResponse('Permission denied', status=403)


        # Get all bookmarks for the user
        bookmarks = Bookmark.objects.filter(user=user)
        # Get the bookmark for the current object
        bookmark = bookmarks.filter(content_type_id=content_type_id, object_id=object_id)

        if bookmark.exists():
            bookmarked = True
        else:
            bookmarked = False


    context = {
        'bookmarked': bookmarked,
        'content_type_id': content_type_id,
        'object_id': object_id,
        'target' : request.htmx.target,
        'size': size
    }
    return render(request, 'components/bookmark.html', context)


from django.conf import settings
import openai
from colorama import Fore, Style

class BloomerpConfigChecker:
    '''Class that checks whether the configuration for bloomerp is correct'''
    settings : dict = settings

    OKAY = 0 # Okay status
    WARNING = 1 # Warning status
    ERROR = 2 # Error status

    def _check_open_ai_key(self) -> tuple[int, str]:
        '''Check if the open ai key is set'''
        openai_key = self.settings.BLOOMERP_SETTINGS.get('OPENAI_API_KEY', None)

        appended_message = 'Without a valid key, the LLM features will not work.'

        if not openai_key:
            return self.WARNING, 'OpenAI key not found in settings' + appended_message
        else:
            # Check if the key is valid
            client = openai.OpenAI(api_key=openai_key)
            try:
                client.models.list()
            except openai.AuthenticationError:
                return self.WARNING, 'Invalid OpenAI key' + appended_message
            else:
                return self.OKAY, 'OpenAI key is valid'

    def _check_login_url(self) -> tuple[int, str]:
        '''Check if the login url is set'''
        login_url = self.settings.LOGIN_URL
        
        if not login_url:
            return self.ERROR, 'Login URL not found in settings'
        
        if login_url != 'login':
            return self.WARNING, 'It is recommended to set the login URL to "login"'
        
        return self.OKAY, 'Login URL is set correctly'

    def check(self) -> bool:
        '''Check the configuration'''
        print(f'{Fore.BLUE}Checking Bloomerp configuration{Style.RESET_ALL}')

        checks = [
            self._check_open_ai_key,
            self._check_login_url
        ]

        for check in checks:
            status, message = check()
            if status == self.OKAY:
                print(f'{Fore.GREEN}OK: {message}{Style.RESET_ALL}')
            elif status == self.WARNING:
                print(f'{Fore.YELLOW}WARNING: {message}{Style.RESET_ALL}')
            else:
                print(f'{Fore.LIGHTRED_EX}ERROR: {message}{Style.RESET_ALL}')